# pylint:disable=unused-private-member,import-outside-toplevel
from __future__ import annotations

import functools
import logging
import os
from typing import TYPE_CHECKING

from angrmanagement.config import Conf, save_config
from angrmanagement.config.config_manager import ENTRIES
from angrmanagement.daemon.client import DaemonClient
from angrmanagement.daemon.url_handler import UrlActionBinaryAware, register_url_action

from .base_plugin import BasePlugin
from .load import load_plugin_descriptions_from_dir, load_plugins_from_file

if TYPE_CHECKING:
    from PySide6.QtGui import QColor

    from angrmanagement.ui.widgets.qblock import QBlock
    from angrmanagement.ui.workspace import Workspace

    from .plugin_description import PluginDescription


log = logging.getLogger(__name__)


class PluginManager:
    """
    The plugin manager can be initialized in two modes:
    - UI mode, where workspace is not None
    - headless mode, where workspace is None, and any plugin that requires workspace will not be initialized

    Plugins can be in two states:
    - Loaded, but inactive. In this, the class is present in .loaded_plugins but there is nothing in .active_plugins
    - Activated. In this, there is an instance of the class in .active_plugins

    ...so this class has functions to transition plugins into and between these states.

    The actual process of excavating a process from the filesystem into a python class ("loading") needs to be dealt
    with before anything touches this class. There are functions to do that in the plugins package but they need to be
    tied to the user's settings related to loading paths and activation. Presently this is split between MainWindow (the
    first-boot autoload part) and the LoadPlugins dialog (the extra loading and tweaking activation)
    """

    def __init__(self, workspace: Workspace | None) -> None:
        self.workspace = workspace
        # should one or both of these be ObjectContainers? I think no since we should be synchronizing on models, not
        # views/controllers. not super clear... that's not a hard and fast rule
        self.loaded_plugins: dict[str, PluginDescription] = {}
        self.active_plugins: dict[str, BasePlugin] = {}

    def discover_and_initialize_plugins(self) -> None:
        os.environ["AM_BUILTIN_PLUGINS"] = os.path.dirname(__file__)
        enabled_plugins = [plugin_.strip() for plugin_ in Conf.enabled_plugins.split(",") if plugin_.strip()]
        for search_dir in Conf.plugin_search_path.split(":"):
            search_dir = os.path.expanduser(search_dir)
            search_dir = os.path.expandvars(search_dir)

            # load plugin descriptions
            dir_and_descs = load_plugin_descriptions_from_dir(search_dir)
            for _, desc in dir_and_descs:
                if desc.shortname in self.loaded_plugins:
                    log.warning(
                        "Plugin shortname conflict: %s and %s have the same shortname %s.",
                        self.loaded_plugins[desc.shortname].plugin_file_path,
                        desc.plugin_file_path,
                        desc.shortname,
                    )
                    continue
                self.loaded_plugins[desc.shortname] = desc

            # activate all enabled plugins
            for plugin_name, desc in self.loaded_plugins.items():
                # don't reactivate already active plugins
                if plugin_name in self.active_plugins:
                    continue

                plugin_conf_key = f"plugin_{desc.name}_enabled"

                # see if the plugin is enabled or not
                if any(
                    (plugin in desc.name or plugin.lower() in desc.shortname.lower()) for plugin in enabled_plugins
                ) and not (hasattr(Conf, plugin_conf_key) and getattr(Conf, plugin_conf_key) is False):
                    # see if we can't load this plugin because headless mode
                    if self.workspace is None and desc.require_workspace:
                        if desc.has_url_actions:
                            basedir = os.path.join(os.path.dirname(desc.plugin_file_path))
                            for plugin_cls in load_plugins_from_file(os.path.join(basedir, desc.entrypoints[0])):
                                if isinstance(plugin_cls, Exception):
                                    log.warning("Exception occurred during plugin loading: %s", plugin_cls)
                                else:
                                    plugin_cls: type[BasePlugin]
                                    for action in plugin_cls.URL_ACTIONS:
                                        register_url_action(action, UrlActionBinaryAware)
                    else:
                        self.activate_plugin_by_name(desc.shortname)

    def verify_plugin_class(self, plugin_cls: type[BasePlugin]):
        if type(plugin_cls) is not type or not issubclass(plugin_cls, BasePlugin):
            raise TypeError("Cannot load a plugin which is not a BasePlugin subclass")
        if hasattr(plugin_cls, f"_{plugin_cls.__name__}__i_hold_this_abstraction_token"):
            raise TypeError("Cannot load an abstract plugin")
        if plugin_cls.REQUIRE_WORKSPACE and self.workspace is None:
            raise RuntimeError("Cannot load plugin %s in headless mode.")

    def activate_plugin_by_name(self, shortname: str) -> None:
        desc = self.loaded_plugins[shortname]
        basedir = os.path.join(os.path.dirname(desc.plugin_file_path))
        for plugin_cls in load_plugins_from_file(os.path.join(basedir, desc.entrypoints[0])):
            if isinstance(plugin_cls, Exception):
                log.warning("Exception occurred during plugin loading: %s", plugin_cls, exc_info=True)
            else:
                self.activate_plugin(desc.shortname, plugin_cls)

    def register_active_plugin(self, shortname: str, plugin_obj: BasePlugin) -> None:
        self.active_plugins[shortname] = plugin_obj
        plugin_cls = plugin_obj.__class__
        plugin_obj.__cached_status_bar_widgets = []
        plugin_obj.__cached_toolbar_actions = []  # a hack, lol. really this could be a mapping on PluginManager but idc
        plugin_obj.__cached_menu_actions = []  # as above

        if self.workspace is not None:
            self._register_status_bar_widgets(plugin_obj)
            self._register_toolbar_actions(plugin_cls, plugin_obj)
            self._register_menu_buttons(plugin_cls, plugin_obj)

            for dview in self.workspace.view_manager.views_by_category["disassembly"]:
                plugin_obj.instrument_disassembly_view(dview)
            for cview in self.workspace.view_manager.views_by_category["pseudocode"]:
                plugin_obj.instrument_code_view(cview)

            self._register_url_actions(plugin_cls, plugin_obj)
            self._register_configuration_entries(plugin_cls)

        for action in plugin_cls.URL_ACTIONS:
            register_url_action(action, UrlActionBinaryAware)

    def activate_plugin(self, shortname: str, plugin_cls: type[BasePlugin]) -> None:
        self.verify_plugin_class(plugin_cls)  # perform the sanity checks
        if self.get_plugin_instance(plugin_cls) is not None:
            return

        try:
            plugin_obj = plugin_cls(self.workspace)
            self.register_active_plugin(shortname, plugin_obj)
        except Exception:  # pylint: disable=broad-except
            log.warning("Plugin %s failed to activate:", plugin_cls.get_display_name(), exc_info=True)
        else:
            log.info("Activated plugin %s", plugin_cls.get_display_name())

    def save_enabled_plugins_to_config(self) -> None:
        # pylint: disable=assigning-non-slot
        Conf.enabled_plugins = ",".join(self.active_plugins.keys())
        save_config()

    def _register_status_bar_widgets(self, plugin: BasePlugin) -> None:
        gen = plugin.status_bar_permanent_widgets()
        if gen is not None:
            for widget in gen:
                self.workspace.main_window.statusBar().addPermanentWidget(widget)
                widget.show()
                plugin.__cached_status_bar_widgets.append(widget)

    def _register_toolbar_actions(self, plugin_cls: type[BasePlugin], plugin: BasePlugin) -> None:
        for idx, (icon, tooltip) in enumerate(plugin_cls.TOOLBAR_BUTTONS):
            from angrmanagement.ui.toolbars.toolbar_action import ToolbarAction

            action = ToolbarAction(
                icon,
                f"plugin {plugin_cls} toolbar {idx}",
                tooltip,
                functools.partial(self._dispatch_single, plugin, BasePlugin.handle_click_toolbar, False, idx),
            )
            plugin.__cached_toolbar_actions.append(action)
            self.workspace._main_window._file_toolbar.add(action)

    def _register_menu_buttons(self, plugin_cls: type[BasePlugin], plugin: BasePlugin) -> None:
        from angrmanagement.ui.menus.menu import MenuEntry, MenuSeparator

        if plugin_cls.MENU_BUTTONS:
            self.workspace._main_window._plugin_menu.add(MenuSeparator())
        for idx, text in enumerate(plugin_cls.MENU_BUTTONS):
            action = MenuEntry(
                text, functools.partial(self._dispatch_single, plugin, BasePlugin.handle_click_menu, False, idx)
            )
            plugin.__cached_menu_actions.append(action)
            self.workspace._main_window._plugin_menu.add(action)

    def _register_url_actions(self, plugin_cls: type[BasePlugin], plugin: BasePlugin) -> None:
        for action in plugin_cls.URL_ACTIONS:
            DaemonClient.register_handler(
                action, functools.partial(self._dispatch_single, plugin, BasePlugin.handle_url_action, False, action)
            )

    @staticmethod
    def _register_configuration_entries(plugin_cls: type[BasePlugin]) -> None:
        new_entries_added = False
        for ent in plugin_cls.CONFIG_ENTRIES:
            if ent not in ENTRIES:
                # if a plugin is disabled and then enabled, the entry may already exist?
                ENTRIES.append(ent)
                new_entries_added = True

        if new_entries_added:
            # reload configuration manager so that it's aware of newly added entries
            Conf.reinterpet()

    def get_plugin_instance_by_name(self, shortname: str) -> BasePlugin | None:
        instances = [plugin for key, plugin in self.active_plugins.items() if key == shortname]
        if not instances:
            return None
        if len(instances) > 1:
            log.error("Somehow there is more than one instance of %s active?", shortname)
        return instances[0]

    def get_plugin_instance(self, plugin_cls: type[BasePlugin]) -> BasePlugin | None:
        instances = [plugin for plugin in self.active_plugins.values() if type(plugin) is plugin_cls]
        if len(instances) == 0:
            return None
        if len(instances) > 1:
            log.error("Somehow there is more than one instance of %s active?", plugin_cls.get_display_name())
        return instances[0]

    def deactivate_plugin_by_name(self, shortname: str) -> None:
        for key, plugin in self.active_plugins.items():
            if key == shortname:
                self.deactivate_plugin(plugin)
                break

    def deactivate_plugin(self, plugin: BasePlugin | type[BasePlugin]) -> None:
        # this method should work on both instances and classes
        if type(plugin) is type:
            plugin = self.get_plugin_instance(plugin)

        if plugin not in self.active_plugins.values():
            return

        for widget in plugin.__cached_status_bar_widgets:
            self.workspace.main_window.statusBar().removeWidget(widget)
        for action in plugin.__cached_toolbar_actions:
            self.workspace._main_window._file_toolbar.remove(action)
        for action in plugin.__cached_menu_actions:
            self.workspace._main_window._plugin_menu.remove(action)

        try:
            plugin.teardown()

        except Exception:  # pylint: disable=broad-except
            log.warning(
                "Plugin %s errored during removal. The UI may be unstable.", plugin.get_display_name(), exc_info=True
            )

        for key in list(self.active_plugins.keys()):
            if self.active_plugins[key] is plugin:
                del self.active_plugins[key]

    #
    # Dispatchers
    #

    def _dispatch(self, func, sensitive, *args):
        for plugin in list(self.active_plugins.values()):
            custom = getattr(plugin, func.__name__)
            if custom.__func__ is not func:
                try:
                    res = custom(*args)
                except Exception as e:  # pylint: disable=broad-except
                    self._handle_error(plugin, func, sensitive, e)
                else:
                    yield res

    def _dispatch_with_plugin(self, func, sensitive, *args):
        for plugin_shortname, plugin in list(self.active_plugins.items()):
            custom = getattr(plugin, func.__name__)
            if custom.__func__ is not func:
                try:
                    res = custom(*args)
                except Exception as e:  # pylint: disable=broad-except
                    self._handle_error(plugin, func, sensitive, e)
                else:
                    yield (plugin_shortname, plugin), res

    def _dispatch_single(self, plugin, func, sensitive, *args):
        custom = getattr(plugin, func.__name__)
        try:
            return custom(*args)
        except Exception as e:  # pylint: disable=broad-except
            self._handle_error(plugin, func, sensitive, e)
            return None

    def _handle_error(self, plugin, func, sensitive, exc) -> None:
        self.workspace.log(f"Plugin {plugin.get_display_name()} errored during {func.__name__}")
        self.workspace.log(exc)
        if sensitive:
            self.workspace.log(f"Deactivating {plugin.get_display_name()} for error during sensitive operation")
            self.deactivate_plugin(plugin)

    def color_insn(self, addr: int, selected, disasm_view) -> QColor | None:
        for res in self._dispatch(BasePlugin.color_insn, True, addr, selected, disasm_view):
            if res is not None:
                return res
        return None

    def color_block(self, addr: int) -> QColor | None:
        for res in self._dispatch(BasePlugin.color_block, True, addr):
            if res is not None:
                return res
        return None

    def color_func(self, func) -> QColor | None:
        for res in self._dispatch(BasePlugin.color_func, True, func):
            if res is not None:
                return res
        return None

    def draw_insn(self, qinsn, painter) -> None:
        for _ in self._dispatch(BasePlugin.draw_insn, True, qinsn, painter):
            pass

    def draw_block(self, qblock, painter) -> None:
        for _ in self._dispatch(BasePlugin.draw_block, True, qblock, painter):
            pass

    def instrument_disassembly_view(self, dview) -> None:
        for _ in self._dispatch(BasePlugin.instrument_disassembly_view, False, dview):
            pass

    def instrument_code_view(self, cview) -> None:
        for _ in self._dispatch(BasePlugin.instrument_code_view, False, cview):
            pass

    def handle_click_insn(self, qinsn, event):
        return any(res for res in self._dispatch(BasePlugin.handle_click_insn, False, qinsn, event))

    def handle_click_block(self, qblock, event):
        return any(res for res in self._dispatch(BasePlugin.handle_click_block, False, qblock, event))

    def handle_raise_view(self, view) -> None:
        for _ in self._dispatch(BasePlugin.handle_raise_view, False, view):
            pass

    def build_qblock_annotations(self, qblock: QBlock):
        for res in self._dispatch(BasePlugin.build_qblock_annotations, False, qblock):
            yield from res

    def build_context_menu_insn(self, insn):
        for res in self._dispatch(BasePlugin.build_context_menu_insn, False, insn):
            yield from res

    def build_context_menu_block(self, block):
        for res in self._dispatch(BasePlugin.build_context_menu_block, False, block):
            yield from res

    def build_context_menu_node(self, node):
        for res in self._dispatch(BasePlugin.build_context_menu_node, False, node):
            yield from res

    def build_context_menu_functions(self, funcs):
        for res in self._dispatch(BasePlugin.build_context_menu_functions, False, funcs):
            yield from res

    def build_context_menu_label(self, addr: int):
        for res in self._dispatch(BasePlugin.build_context_menu_label, False, addr):
            yield from res

    def get_func_column(self, idx: int):
        for plugin in self.active_plugins.values():
            if idx >= len(plugin.FUNC_COLUMNS):
                idx -= len(plugin.FUNC_COLUMNS)
            else:
                return plugin.FUNC_COLUMNS[idx]
        raise IndexError("Not enough columns")

    def count_func_columns(self):
        return sum(len(plugin.FUNC_COLUMNS) for plugin in self.active_plugins.values())

    def extract_func_column(self, func, idx: int):
        for plugin in self.active_plugins.values():
            if idx >= len(plugin.FUNC_COLUMNS):
                idx -= len(plugin.FUNC_COLUMNS)
            else:
                try:
                    return plugin.extract_func_column(func, idx)
                except Exception as e:  # pylint: disable=broad-except
                    # this should really be a "sensitive" operation but like
                    self.workspace.log(e)
                    self.workspace.log("PLEASE FIX YOUR PLUGIN AHHHHHHHHHHHHHHHHH")
                    return 0, ""
        raise IndexError("Not enough columns")

    def step_callback(self, simgr) -> None:
        for _ in self._dispatch(BasePlugin.step_callback, True, simgr):
            pass

    def handle_stack_var_renamed(self, func, offset: int, old_name: str, new_name: str) -> bool:
        for res in self._dispatch(BasePlugin.handle_stack_var_renamed, False, func, offset, old_name, new_name):
            if res:
                return True
        return False

    def handle_stack_var_retyped(self, func, offset: int, old_type, new_type) -> bool:
        for res in self._dispatch(BasePlugin.handle_stack_var_retyped, False, func, offset, old_type, new_type):
            if res:
                return True
        return False

    def handle_func_arg_renamed(self, func, offset: int, old_name: str, new_name: str) -> bool:
        for res in self._dispatch(BasePlugin.handle_func_arg_renamed, False, func, offset, old_name, new_name):
            if res:
                return True
        return False

    def handle_func_arg_retyped(self, func, offset: int, old_type, new_type) -> bool:
        for res in self._dispatch(BasePlugin.handle_func_arg_retyped, False, func, offset, old_type, new_type):
            if res:
                return True
        return False

    def handle_global_var_renamed(self, address, old_name: str, new_name: str) -> bool:
        for res in self._dispatch(BasePlugin.handle_global_var_renamed, False, address, old_name, new_name):
            if res:
                return True
        return False

    def handle_global_var_retyped(self, address, old_type, new_type) -> bool:
        for res in self._dispatch(BasePlugin.handle_global_var_retyped, False, address, old_type, new_type):
            if res:
                return True
        return False

    def handle_other_var_renamed(self, var, old_name: str, new_name: str):
        return any(res for res in self._dispatch(BasePlugin.handle_other_var_renamed, False, var, old_name, new_name))

    def handle_other_var_retyped(self, var, old_type, new_type):
        return any(res for res in self._dispatch(BasePlugin.handle_other_var_retyped, False, var, old_type, new_type))

    def handle_function_renamed(self, func, old_name: str, new_name: str):
        return any(res for res in self._dispatch(BasePlugin.handle_function_renamed, False, func, old_name, new_name))

    def handle_function_retyped(self, func, old_type, new_type):
        return any(res for res in self._dispatch(BasePlugin.handle_global_var_retyped, False, func, old_type, new_type))

    def handle_comment_changed(self, address, old_cmt, new_cmt, created: bool, decomp: bool) -> bool:
        for res in self._dispatch(BasePlugin.handle_comment_changed, False, address, old_cmt, new_cmt, created, decomp):
            if res:
                return True
        return False

    def handle_struct_changed(self, old_struct, new_struct):
        return any(res for res in self._dispatch(BasePlugin.handle_struct_changed, False, old_struct, new_struct))

    def decompile_callback(self, func) -> None:
        for _ in self._dispatch(BasePlugin.decompile_callback, False, func):
            pass

    def handle_project_initialization(self) -> None:
        for _ in self._dispatch(BasePlugin.handle_project_initialization, False):
            pass

    def handle_project_save(self, file_name: str) -> None:
        for _ in self._dispatch(BasePlugin.handle_project_save, False, file_name):
            pass

    def on_workspace_initialized(self, workspace: Workspace) -> None:
        for _ in self._dispatch(BasePlugin.on_workspace_initialized, False, workspace):
            pass

    def angrdb_store_entries(self):
        entries = {}
        for (shortname, _plugin), res in self._dispatch_with_plugin(BasePlugin.angrdb_store_entries, False):
            for result in res:
                if isinstance(result, tuple) and len(result) == 2:
                    key, value = result
                    key = shortname + "___" + key
                    entries[key] = value
        return entries

    def angrdb_load_entries(self, entries) -> None:
        plugin_name_to_plugin = self.active_plugins

        for key, value in entries.items():
            if "___" not in key:
                continue
            splitted = key.split("___")
            if len(splitted) != 2:
                continue
            plugin_shortname, key = splitted
            plugin = plugin_name_to_plugin.get(plugin_shortname, None)
            if plugin is None:
                continue
            # dispatch
            custom = plugin.angrdb_load_entry
            if custom is not BasePlugin.angrdb_load_entry:
                try:
                    custom(key, value)
                except Exception as ex:  # pylint: disable=broad-except
                    self._handle_error(plugin, BasePlugin.angrdb_load_entry, False, ex)

    def optimization_passes(self):
        for plugin in self.active_plugins.values():
            yield from plugin.OPTIMIZATION_PASSES



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.sim_type import normalize_cpp_function_name
from PySide6.QtGui import QColor, Qt
from sortedcontainers import SortedDict

from angrmanagement.plugins.base_plugin import BasePlugin

from .sinks import VulnerabilityType, sink_manager

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class DependencyViewer(BasePlugin):
    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)

        self.covered_blocks = SortedDict()

        self.sink_color = Qt.GlobalColor.yellow

    def color_insn(self, addr: int, selected, disasm_view) -> QColor | None:
        if not selected:
            try:
                block_addr = next(self.covered_blocks.irange(maximum=addr, reverse=True))
            except StopIteration:
                return None
            block_size = self.covered_blocks[block_addr]
            if block_addr <= addr < block_addr + block_size:
                return QColor(0xA5, 0xD0, 0xF3)
        return None

    FUNC_COLUMNS = ("Vuln Sink",)

    def extract_func_column(self, func, idx: int):
        assert idx == 0

        func_name = func.demangled_name
        if "<" in func_name or "{" in func_name:
            func_name = normalize_cpp_function_name(func_name)
            if "(" in func_name:
                # only take function name
                func_name = func_name[: func_name.index("(")]

        vulntype_and_sinks = sink_manager.get_function_sinks(func_name)
        if not vulntype_and_sinks:
            return 0, ""

        vuln_type, sink = vulntype_and_sinks[0]
        return 1, VulnerabilityType.to_string(vuln_type)

    def color_func(self, func) -> QColor | None:
        # test if we have a match
        match = sink_manager.has_function_sink(func.demangled_name)
        if match:
            return self.sink_color

        return None



"""
Command palette dialog for selecting and executing commands.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .palette import PaletteDialog, PaletteItemDelegate, PaletteModel

if TYPE_CHECKING:
    from angrmanagement.logic.commands import Command
    from angrmanagement.ui.workspace import Workspace


class CommandPaletteModel(PaletteModel):
    """
    Data provider for command palette.
    """

    def get_items(self) -> list[Command]:
        return sorted(
            [cmd for cmd in self.workspace.command_manager.get_commands() if cmd.is_visible],
            key=lambda cmd: cmd.caption,
        )

    def get_caption_for_item(self, item: Command) -> str:
        return item.caption


class CommandPaletteDialog(PaletteDialog):
    """
    Dialog for selecting commands.
    """

    def __init__(self, workspace: Workspace, parent=None) -> None:
        super().__init__(CommandPaletteModel(workspace), PaletteItemDelegate(display_icons=False), parent)
        self.setWindowTitle("Command Palette")



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import sys
import unittest

import angr
from common import AngrManagementTestCase, test_location

from angrmanagement.ui.views import DisassemblyView


class TestUndefineCodeRegions(AngrManagementTestCase):
    def test_undefine_code_region_in_disasm_view(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        cfg = main.workspace.main_instance.cfg
        assert cfg.get_any_node(func.addr) is not None

        # load the disassembly view
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        disasm_view.display_function(func)

        # select the instruction at the beginning of the function
        disasm_view.infodock.select_instruction(func.addr)
        assert len(disasm_view.infodock.selected_insns) == 1
        assert disasm_view.infodock.selected_insns == {func.addr}
        # print(len(cfg.graph))

        # undefine at the beginning of the function
        disasm_view.undefine_code()

        main.workspace.job_manager.join_all_jobs()

        # the starting block should no longer be in the graph
        assert cfg.get_any_node(func.addr) is None

        main.workspace.job_manager.join_all_jobs()

        # define at the beginning of the function
        disasm_view.infodock.select_label(func.addr)
        assert disasm_view.infodock.selected_labels == {func.addr}
        # print(len(cfg.graph))
        disasm_view.define_code()

        main.workspace.job_manager.join_all_jobs()
        # print(len(cfg.graph))

        assert cfg.get_any_node(func.addr) is not None


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



# pylint:disable=no-self-use
from __future__ import annotations

import logging
import os
import pickle
import time
import uuid
from functools import partial
from typing import TYPE_CHECKING

import PySide6QtAds as QtAds
from angr.angrdb import AngrDB
from PySide6.QtCore import QEvent, QObject, QSize, Qt, QUrl
from PySide6.QtGui import QDesktopServices, QIcon, QKeySequence, QShortcut, QWindow
from PySide6.QtWidgets import (
    QFileDialog,
    QMainWindow,
    QMessageBox,
    QWidget,
)

from angrmanagement.config import Conf, save_config
from angrmanagement.consts import IMG_LOCATION
from angrmanagement.daemon import daemon_conn, daemon_exists, run_daemon_process
from angrmanagement.daemon.client import ClientService
from angrmanagement.data.jobs import DependencyAnalysisJob
from angrmanagement.data.jobs.loading import LoadAngrDBJob, LoadBinaryJob
from angrmanagement.data.library_docs import LibraryDocs
from angrmanagement.data.signatures import init_flirt_signatures
from angrmanagement.errors import InvalidURLError, UnexpectedStatusCodeError
from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.commands import BasicCommand
from angrmanagement.logic.threads import ExecuteCodeEvent
from angrmanagement.ui.dialogs.progress_dialog import ProgressDialog
from angrmanagement.ui.views import DisassemblyView
from angrmanagement.ui.widgets.qam_status_bar import QAmStatusBar
from angrmanagement.utils.io import download_url, isurl

from .dialogs.about import LoadAboutDialog
from .dialogs.command_palette import CommandPaletteDialog
from .dialogs.goto_palette import GotoPaletteDialog
from .dialogs.load_plugins import LoadPlugins
from .dialogs.new_state import NewState
from .dialogs.preferences import Preferences
from .dialogs.welcome import WelcomeDialog
from .menus.ai_menu import AIMenu
from .menus.analyze_menu import AnalyzeMenu
from .menus.file_menu import FileMenu
from .menus.help_menu import HelpMenu
from .menus.plugin_menu import PluginMenu
from .menus.view_menu import ViewMenu
from .toolbar_manager import ToolbarManager
from .toolbars import DebugToolbar, FeatureMapToolbar, FileToolbar
from .workspace import Workspace

if TYPE_CHECKING:
    from PySide6.QtWidgets import QApplication

    from angrmanagement.data.jobs import Job


_l = logging.getLogger(name=__name__)


class DockShortcutEventFilter(QObject):
    """
    Filter to support shortcuts on floating dock windows that overlap main window registered menu action shortcuts.
    """

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__()
        self._main_window: MainWindow = main_window

    def eventFilter(self, qobject, event) -> bool:
        if event.type() == QEvent.Type.KeyPress and QKeySequence(event.keyCombination()) == QKeySequence(
            "Ctrl+Shift+P"
        ):
            self._main_window.show_command_palette(qobject)
            return True
        return False


class ShiftShiftEventFilter(QObject):
    """
    Filter to catch Shift+Shift key sequence for goto-anything activation.
    """

    activation_key = Qt.Key.Key_Shift
    activation_count: int = 2
    timeout_secs: float = 1

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__()
        self._main_window: MainWindow = main_window
        self._press_count: int = 0
        self._last_press_time: float = 0
        self._did_process_qwindow_event: bool = False

    def eventFilter(self, qobject, event) -> bool:
        # Key Event propagation will begin at QWindow and continue down the widget tree. Use KeyEvent on QWindow to
        # distinguish unique key presses, then intercept KeyEvent at first QWidget the event is propagated to.

        if event.type() == QEvent.Type.KeyPress:
            if isinstance(qobject, QWindow) and qobject.modality() == Qt.WindowModality.NonModal:
                self._did_process_qwindow_event = True
                return False
            if not isinstance(qobject, QWidget) or not self._did_process_qwindow_event:
                return False
            self._did_process_qwindow_event = False

            if event.count() == 1 and event.key() == self.activation_key:
                now = time.time()
                if now - self._last_press_time >= self.timeout_secs:
                    self._press_count = 0
                self._last_press_time = now

                self._press_count += 1
                if self._press_count >= self.activation_count:
                    self._press_count = 0
                    self._main_window.show_goto_palette(qobject)
                    return True

            else:
                self._press_count = 0

        return False


class QExecuteCodeEventReceiver(QObject):
    """
    A class that receives ExecuteCodeEvent and executes the code on the GUI thread.
    """

    def event(self, event, /):
        if not isinstance(event, ExecuteCodeEvent):
            return False
        try:
            event.result = event.execute()
        except Exception as ex:  # pylint:disable=broad-except
            event.exception = ex
            if event.async_:
                _l.exception("Exception occurred in an async job:")
        event.event.set()
        return True


class MainWindow(QMainWindow):
    """
    The main window of angr management.
    """

    def __init__(
        self, app: QApplication | None = None, parent=None, show: bool = True, use_daemon: bool = False
    ) -> None:
        super().__init__(parent)
        self.initialized = False

        icon_location = os.path.join(IMG_LOCATION, "angr.png")
        self.setWindowIcon(QIcon(icon_location))
        self.setWindowTitle("angr management")

        GlobalInfo.main_window = self
        GlobalInfo.event_receiver = QExecuteCodeEventReceiver()
        self.shown_at_start = show

        # initialization
        self.setMinimumSize(QSize(400, 400))

        self.app: QApplication | None = app
        self.workspace: Workspace = None
        self.dock_manager: QtAds.CDockManager
        self._dock_shortcut_event_filter = DockShortcutEventFilter(self)

        self._shift_shift_event_filter = ShiftShiftEventFilter(self)
        if app:
            self.app.installEventFilter(self._shift_shift_event_filter)

        self.toolbar_manager: ToolbarManager = ToolbarManager(self)

        # menus
        self._file_menu = None  # FileMenu
        self._analyze_menu = None
        self._ai_menu = None
        self._view_menu = None
        self._help_menu = None
        self._plugin_menu = None

        self._init_workspace()
        self._init_status_indicators()
        self._init_toolbars()
        self._init_menus()
        self._init_plugins()
        self._init_library_docs()
        self._init_url_scheme_handler()

        self._register_commands()

        self.workspace.plugins.on_workspace_initialized(self)

        self._init_shortcuts()
        init_flirt_signatures()

        self._run_daemon(use_daemon=use_daemon)

        # I'm ready to show off!
        if show:
            self.showMaximized()
            self.windowHandle().screenChanged.connect(self.on_screen_changed)
            self.show()

    def show_welcome_dialog(self) -> None:
        dlg = WelcomeDialog(self)
        dlg.setModal(True)
        dlg.show()

    def sizeHint(self):  # pylint: disable=no-self-use
        return QSize(1200, 800)

    #
    # Properties
    #

    @property
    def caption(self):
        return self.getWindowTitle()

    @caption.setter
    def caption(self, v) -> None:
        self.setWindowTitle(v)

    #
    # Dialogs
    #

    def open_mainfile_dialog(self):
        # pylint: disable=assigning-non-slot
        # https://github.com/PyCQA/pylint/issues/3793
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Open a binary",
            self._get_recent_dir(),
            "All executables (*);;Windows PE files (*.exe);;Core Dumps (*.core);;angr database (*.adb)",
        )
        return file_path

    def open_load_plugins_dialog(self) -> None:
        dlg = LoadPlugins(self.workspace.plugins)
        dlg.setModal(True)
        dlg.exec_()

    def open_newstate_dialog(self) -> None:
        if self.workspace.main_instance.project.am_none:
            QMessageBox.critical(self, "Cannot create new states", "Please open a binary to analyze first.")
            return
        new_state_dialog = NewState(self.workspace, self.workspace.main_instance, parent=self, create_simgr=True)
        new_state_dialog.exec_()

    def open_doc_link(self) -> None:
        QDesktopServices.openUrl(
            QUrl("https://angr-management.readthedocs.io/en/latest/", QUrl.ParsingMode.TolerantMode)
        )

    def open_about_dialog(self) -> None:
        dlg = LoadAboutDialog()
        dlg.exec_()

    #
    # Widgets
    #

    def _init_toolbars(self) -> None:
        for cls in (FileToolbar, DebugToolbar, FeatureMapToolbar):
            self.toolbar_manager.show_toolbar_by_class(cls)

    #
    # Menus
    #

    def _init_menus(self) -> None:
        self._file_menu = FileMenu(self)
        self._analyze_menu = AnalyzeMenu(self)
        self._ai_menu = AIMenu(self)
        self._view_menu = ViewMenu(self)
        self._help_menu = HelpMenu(self)
        self._plugin_menu = PluginMenu(self)

        for path in Conf.recent_files:
            self._file_menu.add_recent(path)

        self.menuBar().addMenu(self._file_menu.qmenu())
        self.menuBar().addMenu(self._view_menu.qmenu())
        self.menuBar().addMenu(self._analyze_menu.qmenu())
        self.menuBar().addMenu(self._ai_menu.qmenu())
        self.menuBar().addMenu(self._plugin_menu.qmenu())
        self.menuBar().addMenu(self._help_menu.qmenu())

    #
    # Workspace
    #

    def _init_workspace(self) -> None:
        """
        Initialize workspace

        :return:    None
        """
        QtAds.CDockManager.setConfigFlags(
            (
                QtAds.CDockManager.DefaultBaseConfig
                | QtAds.CDockManager.OpaqueSplitterResize
                | QtAds.CDockManager.FocusHighlighting
            )
            & ~QtAds.CDockManager.DockAreaHasUndockButton
        )
        self.dock_manager = QtAds.CDockManager(self)
        self.dock_manager.setStyleSheet("")  # Clear stylesheet overrides
        self.dock_manager.setAutoHideConfigFlags(QtAds.CDockManager.DefaultAutoHideConfig)
        self.dock_manager.createSideTabBarWidgets()
        self.setCentralWidget(self.dock_manager)
        wk = Workspace(self)
        self.workspace = wk

        def set_caption(**kwargs) -> None:  # pylint: disable=unused-argument
            if self.workspace.main_instance.project.am_none:
                self.caption = ""
            elif self.workspace.main_instance.project.filename is None:
                self.caption = "Loaded from stream"
            else:
                self.caption = os.path.basename(self.workspace.main_instance.project.filename)

        self.workspace.main_instance.project.am_subscribe(set_caption)

    #
    # Status
    #

    def _init_status_indicators(self):
        """
        Initialize job status indicators, including the status bar and progress dialog.
        """
        self._status_bar = QAmStatusBar(self)
        self._progress_dialog = ProgressDialog(self)
        self._progress_dialog.canceled.connect(self.workspace.job_manager.interrupt_current_job)

        self.workspace.job_manager.job_starting.connect(self._on_job_starting)
        self.workspace.job_manager.job_progressed.connect(self._on_job_progress)
        self.workspace.job_manager.job_exception.connect(self._on_job_finished)
        self.workspace.job_manager.job_finished.connect(self._on_job_finished)

    #
    # Shortcuts
    #

    def _init_shortcuts(self) -> None:
        """
        Initialize shortcuts

        :return:    None
        """

        for i in range(1, 10):
            QShortcut(QKeySequence(f"Alt+{i}"), self, lambda idx=i: self._raise_view(idx - 1))
        QShortcut(QKeySequence("Alt+0"), self, lambda: self._raise_view(9))

        QShortcut(QKeySequence("Ctrl+I"), self, self.workspace.job_manager.interrupt_current_job)

        # Raise the DisassemblyView after everything has initialized
        self._raise_view(0)

        # Toggle exec breakpoint
        QShortcut(QKeySequence(Qt.Key.Key_F2), self, self.workspace.toggle_exec_breakpoint)

        # Single step
        QShortcut(QKeySequence(Qt.Key.Key_F7), self, self.workspace.step_forward)

        # Run
        QShortcut(QKeySequence(Qt.Key.Key_F9), self, self.workspace.continue_forward)

    def init_shortcuts_on_dock(self, dock_widget) -> None:
        """
        Installs an event filter on the dock widget to support floating dock global shortcuts (e.g. command palette).
        """
        dock_widget.installEventFilter(self._dock_shortcut_event_filter)

    #
    # Plugins
    #

    def _init_plugins(self) -> None:
        self.workspace.plugins.discover_and_initialize_plugins()

    #
    # Library docs
    #

    def _init_library_docs(self) -> None:
        GlobalInfo.library_docs = LibraryDocs()
        if Conf.library_docs_root:
            GlobalInfo.library_docs.load_func_docs(Conf.library_docs_root)

    #
    # Daemon
    #

    def _run_daemon(self, use_daemon=None) -> None:
        if use_daemon is None:
            # Load it from the configuration file
            use_daemon = Conf.use_daemon

        if not use_daemon:
            return

        # connect to daemon (if there is one)
        if not daemon_exists():
            print("[+] Starting a new daemon.")
            run_daemon_process()
            time.sleep(0.2)
        else:
            print("[+] Connecting to an existing angr management daemon.")

        while True:
            try:
                GlobalInfo.daemon_conn = daemon_conn(service=ClientService)
            except ConnectionRefusedError:
                print("[-] Connection failed... try again.")
                time.sleep(0.4)
                continue
            print("[+] Connected to daemon.")
            break

        from rpyc import BgServingThread  # pylint:disable=import-outside-toplevel

        _ = BgServingThread(GlobalInfo.daemon_conn)

        GlobalInfo.client_id = str(uuid.uuid4())
        GlobalInfo.daemon_conn.root.register_client(GlobalInfo.client_id)

    #
    # URL scheme handler setup
    #

    def _init_url_scheme_handler(self) -> None:
        if "CI" in os.environ:
            return

        # URL scheme
        from angrmanagement.logic.url_scheme import AngrUrlScheme  # pylint:disable=import-outside-toplevel

        scheme = AngrUrlScheme()
        registered, _ = scheme.is_url_scheme_registered()
        supported = scheme.is_url_scheme_supported()

        if not registered and supported and not Conf.prompted_for_url_scheme_registration:
            btn = QMessageBox.question(
                None,
                "Setting up angr URL scheme",
                'angr URL scheme allows "deep linking" from browsers and other applications '
                "by registering the angr:// protocol to the current user. Do you want to "
                "register it? You may unregister at any "
                "time in Preferences.",
                defaultButton=QMessageBox.StandardButton.Yes,
            )
            if btn == QMessageBox.StandardButton.Yes:
                try:
                    AngrUrlScheme().register_url_scheme()
                except (ValueError, FileNotFoundError) as ex:
                    QMessageBox.warning(
                        None,
                        "Error in registering angr URL scheme",
                        "Failed to register the angr URL scheme.\nThe following exception occurred:\n" + str(ex),
                    )
                    return

            Conf.prompted_for_url_scheme_registration = True
            save_config()

    #
    # Commands
    #

    def _register_commands(self) -> None:
        """
        Register basic window commands.
        """
        self.workspace.command_manager.register_commands(
            [
                BasicCommand(action.__name__, caption, action)
                for caption, action in [
                    ("AI: LLM Refine All", self.llm_refine_all),
                    ("AI: LLM Suggest Variable Names", self.llm_suggest_variable_names),
                    ("AI: LLM Suggest Function Name", self.llm_suggest_function_name),
                    ("AI: LLM Suggest Variable Types", self.llm_suggest_variable_types),
                    ("AI: LLM Summarize Function", self.llm_summarize_function),
                    ("Analyze: Decompile", self.decompile_current_function),
                    ("Analyze: Run Analysis...", self.run_analysis),
                    ("File: Exit", self.quit),
                    ("File: Load a new binary...", self.open_file_button),
                    ("File: Load a new trace...", self.load_trace),
                    ("File: Load angr database...", self.load_database),
                    ("File: Preferences...", self.preferences),
                    ("File: Save angr database as...", self.save_database_as),
                    ("File: Save angr database...", self.save_database),
                    ("File: Save patched binary as...", self.save_patched_binary_as),
                    ("Help: About", self.open_about_dialog),
                    ("Help: Documentation", self.open_doc_link),
                    ("View: Breakpoints", self.workspace.show_breakpoints_view),
                    ("View: Call Explorer", self.workspace.show_call_explorer_view),
                    ("View: Console", self.workspace.show_console_view),
                    ("View: Disassembly (Graph)", self.workspace.show_graph_disassembly_view),
                    ("View: Disassembly (Linear)", self.workspace.show_linear_disassembly_view),
                    ("View: Functions", self.workspace.show_functions_view),
                    ("View: Hex", self.workspace.show_hex_view),
                    ("View: Function Signatures", self.workspace.show_signatures_view),
                    ("View: Jobs", self.workspace.show_jobs_view),
                    ("View: Log", self.workspace.show_log_view),
                    ("View: New Disassembly (Graph)", self.workspace.create_and_show_graph_disassembly_view),
                    ("View: New Disassembly (Linear)", self.workspace.create_and_show_linear_disassembly_view),
                    ("View: New Hex", self.workspace.create_and_show_hex_view),
                    ("View: Patches", self.workspace.show_patches_view),
                    ("View: Proximity", self.workspace.view_proximity_for_current_function),
                    ("View: Pseudocode", self.workspace.show_pseudocode_view),
                    ("View: Registers", self.workspace.show_registers_view),
                    ("View: Stack", self.workspace.show_stack_view),
                    ("View: States", self.workspace.show_states_view),
                    ("View: Strings", self.workspace.show_strings_view),
                    ("View: Symbolic Execution", self.workspace.show_symexec_view),
                    ("View: Trace map", self.workspace.show_trace_map_view),
                    ("View: Traces", self.workspace.show_traces_view),
                    ("View: Types", self.workspace.show_types_view),
                ]
            ]
        )

    #
    # Event
    #

    def closeEvent(self, event) -> None:
        # Ask if the user wants to save things
        if (
            self.workspace.main_instance is not None
            and not self.workspace.main_instance.project.am_none
            and self.shown_at_start
        ):
            msgbox = QMessageBox()
            msgbox.setWindowTitle("Save database")
            msgbox.setText("angr management is about to exit. Do you want to save the database?")
            msgbox.setIcon(QMessageBox.Icon.Question)
            msgbox.setWindowIcon(self.windowIcon())
            msgbox.setStandardButtons(
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Cancel
            )
            msgbox.setDefaultButton(QMessageBox.StandardButton.Yes)
            r = msgbox.exec_()

            if r == QMessageBox.StandardButton.Cancel:
                event.ignore()
                return
            elif r == QMessageBox.StandardButton.Yes:
                save_r = self.save_database()
                if not save_r:
                    # failed to save the database
                    event.ignore()
                    return

        for plugin in list(self.workspace.plugins.active_plugins.values()):
            self.workspace.plugins.deactivate_plugin(plugin)
        self.workspace.job_manager.quit()

        # force-close runtime db
        if (
            self.workspace.main_instance is not None
            and not self.workspace.main_instance.project.am_none
            and "rtdb" in self.workspace.main_instance.project.kb
        ):
            self.workspace.main_instance.kb.rtdb.cleanup()

        event.accept()

    def on_screen_changed(self, screen) -> None:
        """
        When changing from one screen to another, ask disassembly views to refresh in case the DPI is changed.
        """
        self.workspace.current_screen.am_obj = screen
        self.workspace.current_screen.am_event()

    #
    # Actions
    #

    def reload(self) -> None:
        self.workspace.reload()

    def open_file_button(self) -> None:
        file_path = self.open_mainfile_dialog()
        if not file_path:
            return
        self.load_file(file_path)

    def open_trace_file_button(self) -> None:
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Open a trace file",
            Conf.last_used_directory,
            "All files (*);;Trace files (*.trace);;",
        )
        Conf.last_used_directory = os.path.dirname(file_path)
        if not file_path:
            return
        self.load_trace_file(file_path)

    def load_trace_file(self, file_path) -> None:
        if isurl(file_path):
            QMessageBox.critical(
                self,
                "Unsupported Action",
                "Downloading trace files is not yet supported. Please specify a path to a file on disk.",
            )
        else:
            # File
            if os.path.isfile(file_path):
                try:
                    with open(file_path, "rb") as f:
                        loaded_sim_state = pickle.load(f)
                        analysis_params = {
                            "end_state": loaded_sim_state,
                            "start_addr": None,
                            "end_addr": None,
                            "block_addrs": None,
                        }
                        self.workspace.view_data_dependency_graph(analysis_params)
                    self._recent_file(file_path)
                except pickle.PickleError:
                    QMessageBox.critical(
                        self, "Unable to load trace file", "Trace file must contain a serialized SimState."
                    )
            else:
                QMessageBox.critical(
                    self,
                    "File not found",
                    f"angr management cannot open file {file_path}. Please make sure that the file exists.",
                )

    def load_file(self, file_path) -> None:
        if not isurl(file_path):
            # file
            if os.path.isfile(file_path):
                if file_path.endswith(".trace"):
                    self.workspace.load_trace_from_path(file_path)
                    return

                self.workspace.main_instance.binary_path = file_path
                self.workspace.main_instance.original_binary_path = file_path
                if file_path.endswith(".adb"):
                    self._load_database(file_path)
                else:
                    self._recent_file(file_path)
                    self.workspace.job_manager.add_job(LoadBinaryJob(self.workspace.main_instance, file_path))
            else:
                QMessageBox.critical(
                    self,
                    "File not found",
                    f"angr management cannot open file {file_path}. Please make sure that the file exists.",
                )
        else:
            # url
            r = QMessageBox.question(
                self,
                "Downloading a file",
                f"Do you want to download a file from {file_path} and open it in angr management?",
                defaultButton=QMessageBox.StandardButton.Yes,
            )
            if r == QMessageBox.StandardButton.Yes:
                try:
                    target_path = download_url(file_path, parent=self, to_file=True, file_path=None)
                except InvalidURLError:
                    QMessageBox.critical(
                        self, "Downloading failed", "angr management failed to download the file. The URL is invalid."
                    )
                    return
                except UnexpectedStatusCodeError as ex:
                    QMessageBox.critical(
                        self,
                        "Downloading failed",
                        "angr management failed to retrieve the header of the file. "
                        f"The HTTP request returned an unexpected status code {ex.status_code}.",
                    )
                    return

                if target_path:
                    # open the file - now it's a local file
                    self.workspace.main_instance.binary_path = target_path
                    self.workspace.main_instance.original_binary_path = file_path
                    self.load_file(target_path)

    def load_database(self) -> None:
        # Open File window
        file_path, _ = QFileDialog.getOpenFileName(
            self,
            "Load angr database",
            ".",
            "angr databases (*.adb)",
        )

        if not file_path:
            return

        self._load_database(file_path)

    def save_database(self):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return True

        if self.workspace.main_instance.database_path is None:
            return self.save_database_as()
        else:
            return self._save_database(self.workspace.main_instance.database_path)

    def save_database_as(self):
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return False

        default_database_path = self.workspace.main_instance.database_path
        if default_database_path is None:
            default_database_path = os.path.normpath(self.workspace.main_instance.project.filename) + ".adb"

        # Open File window
        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save angr database",
            default_database_path,
            "angr databases (*.adb)",
        )

        if not file_path:
            return False

        if not file_path.endswith(".adb"):
            file_path = file_path + ".adb"

        return self._save_database(file_path)

    def save_patched_binary_as(self) -> None:
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return

        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save patched binary as...",
            self.workspace.main_instance.project.loader.main_object.binary
            + ".patched",  # FIXME: this will not work if we are loading from an angrdb
            "Any file (*)",
        )

        if file_path:
            b = self.workspace.main_instance.project.kb.patches.apply_patches_to_binary()
            with open(file_path, "wb") as f:
                f.write(b)

    def load_trace(self) -> None:
        file_path, _ = QFileDialog.getOpenFileName(self, "Load trace", ".", "bintrace (*.trace)")
        if not file_path:
            return
        self.workspace.load_trace_from_path(file_path)

    def preferences(self) -> None:
        pref = Preferences(self.workspace, parent=self)
        pref.exec_()

    def quit(self) -> None:
        self.close()

    def run_variable_recovery(self) -> None:
        self.workspace._get_or_create_view("disassembly", DisassemblyView).variable_recovery_flavor = "accurate"

    def run_induction_variable_analysis(self) -> None:
        self.workspace._get_or_create_view("disassembly", DisassemblyView).run_induction_variable_analysis()

    def run_dependency_analysis(self, func_addr: int | None = None, func_arg_idx: int | None = None) -> None:
        if self.workspace is None or self.workspace.main_instance is None:
            return
        dep_analysis_job = DependencyAnalysisJob(
            self.workspace.main_instance, func_addr=func_addr, func_arg_idx=func_arg_idx
        )
        self.workspace.job_manager.add_job(dep_analysis_job)

    def run_analysis(self) -> None:
        if self.workspace:
            self.workspace.run_analysis()

    def decompile_current_function(self) -> None:
        if self.workspace is not None:
            self.workspace.decompile_current_function()

    def llm_refine_all(self) -> None:
        if self.workspace is not None:
            self.workspace.llm_refine_all()

    def llm_suggest_variable_names(self) -> None:
        if self.workspace is not None:
            self.workspace.llm_suggest_variable_names()

    def llm_suggest_function_name(self) -> None:
        if self.workspace is not None:
            self.workspace.llm_suggest_function_name()

    def llm_suggest_variable_types(self) -> None:
        if self.workspace is not None:
            self.workspace.llm_suggest_variable_types()

    def llm_summarize_function(self) -> None:
        if self.workspace is not None:
            self.workspace.llm_summarize_function()

    def view_proximity_for_current_function(self) -> None:
        if self.workspace is not None:
            self.workspace.view_proximity_for_current_function()

    def show_command_palette(self, parent=None) -> None:
        dlg = CommandPaletteDialog(self.workspace, parent=(parent or self))
        dlg.setModal(True)
        dlg.exec_()
        if dlg.selected_item:
            dlg.selected_item.run()

    def show_goto_palette(self, parent=None) -> None:
        dlg = GotoPaletteDialog(self.workspace, parent=(parent or self))
        dlg.setModal(True)
        dlg.exec_()
        if dlg.selected_item:
            self.workspace.jump_to(dlg.selected_item.addr)

    #
    # Other public methods
    #

    def bring_to_front(self) -> None:
        self.setWindowState((self.windowState() & ~Qt.WindowState.WindowMinimized) | Qt.WindowState.WindowActive)
        self.activateWindow()
        self.raise_()

    #
    # Private methods
    #

    def _recent_file(self, file_path) -> None:
        file_path = os.path.abspath(file_path)
        self._file_menu.add_recent(file_path)
        Conf.recent_file(file_path)
        save_config()

    def _get_recent_dir(self) -> str:
        if Conf.recent_files:
            recent_dir = os.path.dirname(Conf.recent_files[-1])
            if os.path.isdir(recent_dir):
                return recent_dir
        return ""

    def _load_database(self, file_path: str) -> None:
        other_kbs = {}
        extra_info = {}

        job = LoadAngrDBJob(
            self.workspace.main_instance,
            file_path,
            ["global", "pseudocode_variable_kb"],
            other_kbs=other_kbs,
            extra_info=extra_info,
        )
        # TODO: make the job return what the callback wants
        job._on_finish = partial(self._on_load_database_finished, job)
        self.workspace.job_manager.add_job(job)

    def _on_load_database_finished(self, job: LoadAngrDBJob, *args, **kwargs) -> None:  # pylint:disable=unused-argument
        proj = job.project

        if proj is None:
            return

        self._recent_file(job.file_path)

        cfg = proj.kb.cfgs["CFGFast"]
        cfb = proj.analyses.CFB()  # it will load functions from kb

        self.workspace.main_instance.database_path = job.file_path

        self.workspace.main_instance._reset_containers()
        self.workspace.main_instance.project = proj
        self.workspace.main_instance.cfg = cfg
        self.workspace.main_instance.cfb = cfb
        if "pseudocode_variable_kb" in job.other_kbs:
            self.workspace.main_instance.pseudocode_variable_kb = job.other_kbs["pseudocode_variable_kb"]
        else:
            self.workspace.main_instance.initialize_pseudocode_variable_kb()
        self.workspace.main_instance.project.am_event(initialized=True)

        # trigger callbacks
        self.workspace.reload()
        self.workspace.main_instance.cfb.am_event()
        self.workspace.main_instance.cfg.am_event()
        self.workspace.on_cfg_generated()
        self.workspace.plugins.angrdb_load_entries(job.extra_info)

    def _save_database(self, file_path) -> bool:
        if self.workspace.main_instance is None or self.workspace.main_instance.project.am_none:
            return False

        self.workspace.plugins.handle_project_save(file_path)

        angrdb = AngrDB(project=self.workspace.main_instance.project)
        extra_info = self.workspace.plugins.angrdb_store_entries()
        angrdb.dump(
            file_path,
            kbs=[
                self.workspace.main_instance.kb,
                self.workspace.main_instance.pseudocode_variable_kb,
            ],
            extra_info=extra_info,
        )

        self.workspace.main_instance.database_path = file_path
        return True

    def _raise_view(self, idx: int) -> None:
        """
        Raise idx'th view in the dock manager
        """
        try:
            dock = self.dock_manager.dockWidgets()[idx]
        except IndexError:
            return
        dock.raise_()

    def _on_job_starting(self, job: Job) -> None:
        text, progress = "Working...", 0.0
        self._status_bar.progress(text, progress, True)
        if job.blocking:
            self._progress_dialog.setLabelText(text)
            self._progress_dialog.setValue(round(progress))

    def _on_job_progress(self, job: Job, percentage: float, text: str = "") -> None:
        text = f"{job.name}: {text}" if text else job.name
        self._status_bar.progress(text, percentage)
        if job.blocking:
            self._progress_dialog.setLabelText(text)
            self._progress_dialog.setValue(round(percentage))

    def _on_job_finished(self, job: Job) -> None:  # pylint:disable=unused-argument
        self._status_bar.progress_done()
        self._progress_dialog.reset()



from __future__ import annotations

import unittest

from common import AngrManagementTestCase, ProjectOpenTestCase

from angrmanagement.ui.views import (
    BreakpointsView,
    CallExplorerView,
    CodeView,
    ConsoleView,
    DataDepView,
    DependencyView,
    DisassemblyView,
    FunctionsView,
    HexView,
    JobsView,
    LogView,
    PatchesView,
    ProximityView,
    RegistersView,
    StackView,
    StringsView,
    SymexecView,
    TraceMapView,
    TracesView,
    TypesView,
)


class TestViewsOpen(AngrManagementTestCase):
    """Tests that all views open, without first opening a project."""

    def test_breakpoints_view(self):
        self.main.workspace.show_view("breakpoints", BreakpointsView)

    def test_call_explorer_view(self):
        self.main.workspace.show_view("call_explorer", CallExplorerView)

    def test_code_view(self):
        self.main.workspace.show_view("code", CodeView)

    def test_console_view(self):
        self.main.workspace.show_view("console", ConsoleView)

    def test_data_dep_view(self):
        self.main.workspace.show_view("data_dep", DataDepView)

    def test_dependency_view(self):
        self.main.workspace.show_view("dependency", DependencyView)

    def test_disassembly_view(self):
        self.main.workspace.show_view("disassembly", DisassemblyView)

    def test_functions_view(self):
        self.main.workspace.show_view("functions", FunctionsView)

    def test_hex_view(self):
        self.main.workspace.show_view("hex", HexView)

    def test_jobs_view(self):
        self.main.workspace.show_view("jobs", JobsView)

    def test_log_view(self):
        self.main.workspace.show_view("log", LogView)

    def test_patches_view(self):
        self.main.workspace.show_view("patches", PatchesView)

    def test_proximity_view(self):
        self.main.workspace.show_view("proximity", ProximityView)

    def test_registers_view(self):
        self.main.workspace.show_view("registers", RegistersView)

    def test_stack_view(self):
        self.main.workspace.show_view("stack", StackView)

    def test_strings_view(self):
        self.main.workspace.show_view("strings", StringsView)

    def test_symexec_view(self):
        self.main.workspace.show_view("symexec", SymexecView)

    def test_trace_map_view(self):
        self.main.workspace.show_view("trace_map", TraceMapView)

    def test_traces_view(self):
        self.main.workspace.show_view("traces", TracesView)

    def test_types_view(self):
        self.main.workspace.show_view("types", TypesView)


class TestViewsOpenWithProject(ProjectOpenTestCase, TestViewsOpen):
    """Tests that all views open, after opening a project."""


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING

from cachetools import LRUCache

if TYPE_CHECKING:
    from collections.abc import Callable


class SmartLRUCache(LRUCache):
    """
    A smart LRU cache that calls an eviction function when an item is evicted from the cache.

    This is based the SmartLRUCache in claripy. Because we may make claripy optional in the future, I decide to make a
    copy of this class in angr management instead.
    """

    def __init__(self, maxsize, getsizeof=None, evict: Callable | None = None):
        LRUCache.__init__(self, maxsize, getsizeof=getsizeof)
        self._evict: Callable | None = evict

    def popitem(self):
        key, val = LRUCache.popitem(self)
        if self._evict is not None:
            self._evict(key, val)
        return key, val



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any

from PySide6.QtCore import QPointF, QRectF, Qt
from PySide6.QtGui import QPainter, QTextCursor, QTextDocument
from PySide6.QtWidgets import QGraphicsSceneMouseEvent, QGraphicsSimpleTextItem

from angrmanagement.config import Conf, ConfigurationManager

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.disassembly.info_dock import InfoDock
    from angrmanagement.ui.views import DisassemblyView

    from .block_code_objects import BlockTreeNode


class QBlockCodeSelectionMode(Enum):
    """
    Defines the selection mode of this QBlockCode object.
    """

    SELECT_INSTRUCTION = 1
    INVOKE_OBJ = 2
    NOOP = 3


class QBlockCode(QCachedGraphicsItem):
    """
    Top-level code widget for a selection of text. Will construct an AST using
    QBlockCodeObj, mirroring the structure associated with the target object.
    This text is then rendered using a QTextDocument, with appropriate styles
    applied to it. Interaction events will be propagated to corresponding
    objects.
    """

    GRAPH_ADDR_SPACING = 20

    addr: int | None
    _addr_str: str
    obj: BlockTreeNode
    _config: ConfigurationManager
    disasm_view: DisassemblyView
    infodock: InfoDock
    parent: Any

    def __init__(
        self,
        addr: int | None,
        obj: BlockTreeNode,
        config: ConfigurationManager,
        instance: Instance,
        infodock: InfoDock,
        parent: Any = None,
        selection_mode: QBlockCodeSelectionMode = QBlockCodeSelectionMode.SELECT_INSTRUCTION,
    ) -> None:
        super().__init__(parent=parent)
        self.addr = addr
        self._addr_str = "" if self.addr is None else f"{self.addr:08x}"
        self.obj = obj
        self._width = 0
        self._height = 0
        self._config = config
        self.parent = parent
        self.instance = instance
        self.infodock = infodock
        self._selection_mode = selection_mode
        self._disasm_view = infodock.disasm_view
        self._qtextdoc = QTextDocument()
        self._qtextdoc.setDefaultFont(self._config.disasm_font)
        self._qtextdoc.setDocumentMargin(0)

        self._addr_item = None
        if self.obj.display_address:
            self._addr_item = QGraphicsSimpleTextItem(self._addr_str, self)
            self._addr_item.setBrush(Conf.disasm_view_node_address_color)
            self._addr_item.setFont(Conf.disasm_font)

        self.update_document()
        self.setToolTip("Address: " + self._addr_str)

        self.refresh()

    def refresh(self) -> None:
        if self._addr_item is not None:
            self._addr_item.setVisible(self._disasm_view.show_address)
        self._layout_items_and_update_size()

    def update_document(self) -> None:
        self._qtextdoc.clear()
        cur = QTextCursor(self._qtextdoc)
        self.obj.render_to_doc(cur)

    def paint(self, painter, option, widget=None) -> None:  # pylint: disable=unused-argument
        self.update_document()
        painter.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)
        painter.setFont(self._config.disasm_font)

        if self.should_highlight():
            highlight_color = Conf.disasm_view_node_instruction_selected_background_color
            painter.setBrush(highlight_color)
            painter.setPen(highlight_color)
            painter.drawRect(0, 0, self.width, self.height)

        y = self.obj.top_margin_lines * Conf.disasm_font_height
        x = 0
        if self._disasm_view.show_address and self.obj.display_address:
            x += self._addr_item.boundingRect().width() + self.GRAPH_ADDR_SPACING

        painter.translate(QPointF(x, y))
        self._qtextdoc.drawContents(painter)

    #
    # Event handlers
    #

    def get_obj_for_mouse_event(self, event: QGraphicsSceneMouseEvent) -> BlockTreeNode | None:
        p = event.pos()

        if self._disasm_view.show_address and self.obj.display_address:
            offset = self._addr_item.boundingRect().width() + self.GRAPH_ADDR_SPACING
            p.setX(p.x() - offset)

        if self.obj.top_margin_lines > 0:
            p.setY(p.y() - self.obj.top_margin_lines * Conf.disasm_font_height)

        if p.x() >= 0:
            hitpos = self._qtextdoc.documentLayout().hitTest(p, Qt.HitTestAccuracy.ExactHit)
            if hitpos >= 0:
                return self.obj.get_hit_obj(hitpos)

        return None

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        if (
            event.button() == Qt.MouseButton.LeftButton
            and self._selection_mode == QBlockCodeSelectionMode.SELECT_INSTRUCTION
            and self.addr is not None
        ):
            self.infodock.select_instruction(self.addr)

        obj = self.get_obj_for_mouse_event(event)
        if obj is not None:
            obj.mousePressEvent(event)

    def mouseDoubleClickEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        obj = self.get_obj_for_mouse_event(event)
        if obj is not None:
            obj.mouseDoubleClickEvent(event)

    def should_highlight(self) -> bool:
        match self._selection_mode:
            case QBlockCodeSelectionMode.SELECT_INSTRUCTION:
                if self.addr is not None and self.infodock.is_instruction_selected(self.addr):
                    return True
                if self.obj is not None:
                    return self.obj.should_highlight() or self.obj.should_highlight_line
            case QBlockCodeSelectionMode.INVOKE_OBJ:
                if self.obj is not None:
                    return self.obj.should_highlight() or self.obj.should_highlight_line
            case QBlockCodeSelectionMode.NOOP:
                return False
        return False

    #
    # Private methods
    #

    def _layout_items_and_update_size(self) -> None:
        self.update_document()

        x, y = 0, self.obj.top_margin_lines * Conf.disasm_font_height
        if self._disasm_view.show_address and self.obj.display_address:
            self._addr_item.setPos(x, y)
            x += self._addr_item.boundingRect().width() + self.GRAPH_ADDR_SPACING

        x += self._qtextdoc.size().width()
        y += self._qtextdoc.size().height()
        y += self.obj.bottom_margin_lines * Conf.disasm_font_height
        self._width = x
        self._height = y
        self.recalculate_size()

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QRectF
from PySide6.QtWidgets import QGraphicsSimpleTextItem

from angrmanagement.config import Conf

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance


class QAlignmentBlock(QCachedGraphicsItem):
    """Renders an alignment function as a single summary line in the linear disassembly view."""

    SPACING = 20

    def __init__(self, instance: Instance, addr: int, size: int, disasm_view, parent=None) -> None:
        super().__init__(parent=parent)

        self.instance = instance
        self.addr = addr
        self.size = size
        self.disasm_view = disasm_view

        self._width = 0
        self._height = 0

        self._addr_item: QGraphicsSimpleTextItem | None = None
        self._content_item: QGraphicsSimpleTextItem | None = None

        self._init_widgets()

    #
    # Public methods
    #

    def refresh(self) -> None:
        self._layout_items_and_update_size()

    def setVisible(self, visible) -> None:
        super().setVisible(visible)
        self._addr_item.setVisible(visible and self.disasm_view.show_address)
        self._content_item.setVisible(visible)

    def paint(self, painter, option, widget=None) -> None:  # pylint: disable=unused-argument
        pass

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)

    def remove_children_from_scene(self):
        """
        Remove this item and all its children from the scene.
        """
        scene = self.scene()
        if scene is None:
            return

        if self._addr_item is not None:
            scene.removeItem(self._addr_item)
            self._addr_item = None

        if self._content_item is not None:
            scene.removeItem(self._content_item)
            self._content_item = None

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self._addr_item = QGraphicsSimpleTextItem(f"{self.addr:08x}", self)
        self._addr_item.setBrush(Conf.disasm_view_node_address_color)
        self._addr_item.setFont(Conf.disasm_font)

        self._content_item = QGraphicsSimpleTextItem(f"[alignment_chunk]: {self.size:#x} bytes", self)
        self._content_item.setBrush(Conf.disasm_view_alignment_color)
        self._content_item.setFont(Conf.disasm_font)

        self._layout_items_and_update_size()

    def _layout_items_and_update_size(self) -> None:
        x, y = 0, Conf.disasm_font_height

        if self.disasm_view.show_address:
            self._addr_item.setVisible(True)
            self._addr_item.setPos(x, y)
            x += self._addr_item.boundingRect().width()
            x += self.SPACING
        else:
            self._addr_item.setVisible(False)

        self._content_item.setPos(x, y)

        self._width = x + self._content_item.boundingRect().width()
        self._height = y + self._content_item.boundingRect().height()

        self.recalculate_size()



"""
Test cases for Command Palette.
"""
# pylint: disable=no-self-use

from __future__ import annotations

import unittest
from unittest.mock import MagicMock, patch

from common import AngrManagementTestCase
from PySide6.QtWidgets import QApplication

from angrmanagement.logic.commands import BasicCommand, Command, ViewCommand
from angrmanagement.logic.commands.command_manager import CommandManager
from angrmanagement.ui.dialogs.command_palette import (
    CommandPaletteDialog,
    CommandPaletteModel,
)


class DummyView:
    """Lightweight dummy view class for testing ViewCommand without UI dependencies."""


class TestCommand(unittest.TestCase):
    """Test Command base class functionality."""

    def test_command_name_default(self):
        """Test that command name defaults to class name."""

        class TestCmd(Command):
            """Test command."""

        cmd = TestCmd()
        assert cmd.name == "TestCmd"

    def test_command_name_override(self):
        """Test that command name can be overridden."""

        class TestCmd(Command):
            """Test command."""

            _name = "custom_name"

        cmd = TestCmd()
        assert cmd.name == "custom_name"

    def test_command_caption_default(self):
        """Test that command caption defaults to name."""

        class TestCmd(Command):
            """Test command."""

        cmd = TestCmd()
        assert cmd.caption == "TestCmd"

    def test_command_caption_override(self):
        """Test that command caption can be overridden."""

        class TestCmd(Command):
            """Test command."""

            _name = "test_name"
            _caption = "Test Caption"

        cmd = TestCmd()
        assert cmd.caption == "Test Caption"

    def test_command_is_visible_default(self):
        """Test that command is visible by default."""

        class TestCmd(Command):
            """Test command."""

        cmd = TestCmd()
        assert cmd.is_visible is True


class TestBasicCommand(unittest.TestCase):
    """Test BasicCommand functionality."""

    def test_basic_command_initialization(self):
        """Test that BasicCommand initializes correctly."""
        action = MagicMock()
        cmd = BasicCommand("test_cmd", "Test Command", action)

        assert cmd.name == "test_cmd"
        assert cmd.caption == "Test Command"

    def test_basic_command_run(self):
        """Test that BasicCommand.run() calls the action."""
        action = MagicMock()
        cmd = BasicCommand("test_cmd", "Test Command", action)

        cmd.run()
        action.assert_called_once()


class TestViewCommand(unittest.TestCase):
    """Test ViewCommand functionality."""

    def setUp(self):
        self.mock_workspace = MagicMock()
        self.mock_view_manager = MagicMock()
        self.mock_workspace.view_manager = self.mock_view_manager

    def test_view_command_initialization(self):
        """Test that ViewCommand initializes correctly."""
        action = MagicMock()
        cmd = ViewCommand("test_cmd", "Test Command", action, DummyView, self.mock_workspace)  # type: ignore[arg-type]

        assert cmd.name == "test_cmd"
        assert cmd.caption == "Test Command"

    def test_view_command_is_visible_with_matching_view(self):
        """Test that ViewCommand is visible when matching view is focused."""
        action = MagicMock()
        cmd = ViewCommand("test_cmd", "Test Command", action, DummyView, self.mock_workspace)  # type: ignore[arg-type]

        mock_view = MagicMock(spec=DummyView)
        self.mock_view_manager.most_recently_focused_view = mock_view

        assert cmd.is_visible is True

    def test_view_command_is_not_visible_without_matching_view(self):
        """Test that ViewCommand is not visible when no matching view is focused."""
        action = MagicMock()
        cmd = ViewCommand("test_cmd", "Test Command", action, DummyView, self.mock_workspace)  # type: ignore[arg-type]

        self.mock_view_manager.most_recently_focused_view = None

        assert cmd.is_visible is False

    def test_view_command_run_with_matching_view(self):
        """Test that ViewCommand.run() calls action with matching view."""
        action = MagicMock()
        cmd = ViewCommand("test_cmd", "Test Command", action, DummyView, self.mock_workspace)  # type: ignore[arg-type]

        mock_view = MagicMock(spec=DummyView)
        self.mock_view_manager.most_recently_focused_view = mock_view

        cmd.run()
        action.assert_called_once_with(mock_view)

    def test_view_command_run_without_matching_view(self):
        """Test that ViewCommand.run() does nothing when no matching view is focused."""
        action = MagicMock()
        cmd = ViewCommand("test_cmd", "Test Command", action, DummyView, self.mock_workspace)  # type: ignore[arg-type]

        self.mock_view_manager.most_recently_focused_view = None

        cmd.run()
        action.assert_not_called()


class TestCommandPaletteModel(unittest.TestCase):
    """Test CommandPaletteModel functionality."""

    def setUp(self):
        self.mock_workspace = MagicMock()
        self.command_manager = CommandManager()
        self.mock_workspace.command_manager = self.command_manager
        self.model = CommandPaletteModel(self.mock_workspace)

    def test_model_initialization(self):
        """Test that CommandPaletteModel initializes correctly."""
        assert self.model.workspace == self.mock_workspace
        assert isinstance(self.model._available_items, list)
        assert isinstance(self.model._filtered_items, list)

    def test_get_items_returns_sorted_visible_commands(self):
        """Test that get_items returns only visible commands sorted by caption."""
        cmd1 = BasicCommand("cmd1", "Z Command", MagicMock())
        cmd2 = BasicCommand("cmd2", "A Command", MagicMock())

        class InvisibleCommand(Command):
            """Test command that is not visible."""

            _caption = "Invisible"

            @property
            def is_visible(self):
                return False

        cmd3 = InvisibleCommand()

        self.command_manager.register_command(cmd1)
        self.command_manager.register_command(cmd2)
        self.command_manager.register_command(cmd3)

        model = CommandPaletteModel(self.mock_workspace)
        items = model.get_items()

        assert cmd3 not in items
        visible_commands = [item for item in items if item in (cmd1, cmd2)]
        assert len(visible_commands) == 2
        captions = [model.get_caption_for_item(item) for item in visible_commands]
        assert captions == sorted(captions)

    def test_get_caption_for_item(self):
        """Test that get_caption_for_item returns command caption."""
        cmd = BasicCommand("test_cmd", "Test Command", MagicMock())
        caption = self.model.get_caption_for_item(cmd)
        assert caption == "Test Command"


class TestCommandPaletteDialog(AngrManagementTestCase):
    """Test CommandPaletteDialog functionality."""

    def setUp(self):
        super().setUp()
        self.mock_workspace = MagicMock()
        self.command_manager = CommandManager()
        self.mock_workspace.command_manager = self.command_manager
        self.dialog = CommandPaletteDialog(self.mock_workspace, parent=self.main)

    def tearDown(self):
        if hasattr(self, "dialog"):
            self.dialog.close()
            del self.dialog
        super().tearDown()

    def test_dialog_initialization(self):
        """Test that CommandPaletteDialog initializes correctly."""
        assert isinstance(self.dialog._model, CommandPaletteModel)
        assert self.dialog.windowTitle() == "Command Palette"

    def test_dialog_uses_workspace(self):
        """Test that dialog uses the provided workspace."""
        assert self.dialog._model.workspace == self.mock_workspace

    def test_dialog_shows_all_commands(self):
        """Test that dialog shows all visible commands."""
        cmd1 = BasicCommand("cmd1", "Test Command 1", MagicMock())
        cmd2 = BasicCommand("cmd2", "Test Command 2", MagicMock())
        self.command_manager.register_command(cmd1)
        self.command_manager.register_command(cmd2)

        dialog = CommandPaletteDialog(self.mock_workspace, parent=self.main)
        model = dialog._model

        items = [model.data(model.index(i, 0)) for i in range(model.rowCount())]
        assert cmd1 in items or cmd2 in items

        dialog.close()

    def test_dialog_filters_commands_by_text(self):
        """Test that dialog filters commands by query text."""
        cmd1 = BasicCommand("cmd1", "Unique Test Command", MagicMock())
        cmd2 = BasicCommand("cmd2", "Other Command", MagicMock())
        self.command_manager.register_command(cmd1)
        self.command_manager.register_command(cmd2)

        dialog = CommandPaletteDialog(self.mock_workspace, parent=self.main)

        dialog._query.setText("Unique")
        QApplication.processEvents()

        items = [dialog._model.data(dialog._model.index(i, 0)) for i in range(dialog._model.rowCount())]
        captions = [dialog._model.get_caption_for_item(item) for item in items if item is not None]

        assert any("Unique" in caption for caption in captions)

        dialog.close()


class TestCommandPaletteIntegration(AngrManagementTestCase):
    """Test Command Palette integration with MainWindow."""

    def test_show_command_palette_creates_dialog(self):
        """Test that show_command_palette creates and shows CommandPaletteDialog."""
        with (
            patch("angrmanagement.ui.main_window.CommandPaletteDialog") as mock_dialog_class,
            patch.object(self.main, "workspace"),
        ):
            mock_dialog = MagicMock()
            mock_dialog_class.return_value = mock_dialog
            mock_dialog.selected_item = None

            self.main.show_command_palette()

            mock_dialog_class.assert_called_once_with(self.main.workspace, parent=self.main)
            mock_dialog.setModal.assert_called_once_with(True)
            mock_dialog.exec_.assert_called_once()

    def test_show_command_palette_runs_selected_command(self):
        """Test that show_command_palette runs the selected command."""
        with (
            patch("angrmanagement.ui.main_window.CommandPaletteDialog") as mock_dialog_class,
            patch.object(self.main, "workspace"),
        ):
            mock_command = MagicMock()
            mock_dialog = MagicMock()
            mock_dialog_class.return_value = mock_dialog
            mock_dialog.selected_item = mock_command

            self.main.show_command_palette()

            mock_command.run.assert_called_once()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class SimgrStepJob(InstanceJob):
    """A job that runs the step method of the simulation manager."""

    def __init__(self, instance: Instance, simgr, until_branch: bool = False, step_callback=None) -> None:
        super().__init__("Simulation manager stepping", instance, on_finish=self._fire_completion_event)
        self._simgr = simgr
        self._until_branch = until_branch
        self._step_callback = step_callback

    def run(self, _: JobContext):
        if self._until_branch:
            orig_len = len(self._simgr.active)
            if orig_len > 0:
                while len(self._simgr.active) == orig_len:
                    self._simgr.step(step_func=self._step_callback)
                    self._simgr.prune()
        else:
            self._simgr.step(step_func=self._step_callback, num_inst=1)
            self._simgr.prune()

        return self._simgr

    def _fire_completion_event(self, result) -> None:
        self._simgr.am_event(src="job_done", job="step", result=result)

    def __repr__(self) -> str:
        if self._until_branch:
            return f"Stepping {self._simgr!r} until branch"
        else:
            return f"Stepping {self._simgr!r}"



from __future__ import annotations

from collections import defaultdict
from inspect import isfunction
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtGui import QContextMenuEvent, QCursor
from PySide6.QtWidgets import QAbstractItemView, QInputDialog, QMenu, QMessageBox, QTreeWidget, QTreeWidgetItem

if TYPE_CHECKING:
    from angr import SimState


class SimgrViewerAbstractTreeItem(QTreeWidgetItem):
    def handle_context_menu_event(self, event: QContextMenuEvent):
        """Handles right-click actions on the specific QTreeWidgetItem that was clicked in the view"""
        raise NotImplementedError


class StashTreeItem(SimgrViewerAbstractTreeItem):
    def __init__(self, stash_name: str, simgr_viewer) -> None:
        self.simgr_viewer = simgr_viewer
        self.stash_name = stash_name
        super().__init__(simgr_viewer)
        self.setFlags(self.flags() & ~Qt.ItemFlag.ItemIsSelectable)
        self.refresh()

    def __iter__(self):
        for i in range(self.childCount()):
            yield self.child(i)

    @property
    def states(self):
        return self.simgr_viewer.simgr.stashes[self.stash_name]

    def refresh(self) -> None:
        self.takeChildren()
        for state in self.simgr_viewer.simgr.stashes[self.stash_name]:
            if self.stash_name == "errored" and getattr(state, "state", None):
                state = state.state
            self.addChild(StateTreeItem(state, self.simgr_viewer))
        self.setText(0, f"{self.stash_name} ({len(self.states)})")

    def handle_context_menu_event(self, event) -> None:
        menu = QMenu()
        menu.addAction("Copy states", self.copy_states)
        menu.addAction("Cut states", self.cut_states)
        if self.simgr_viewer.state_clipboard:
            plural = ""
            if len(self.simgr_viewer.state_clipboard) > 1:
                plural += "s"
            menu.addAction("Paste state" + plural, self.paste_states)
        menu.addAction("Delete stash", self.delete_stash)
        if self.stash_name != "active":
            menu.addAction("Move states to here", self.move_states)
        menu.exec_(QCursor.pos())

    def copy_states(self) -> None:
        self.simgr_viewer.state_clipboard = [s.state for s in self]
        self.refresh()

    def cut_states(self) -> None:
        self.simgr_viewer.state_clipboard = [s.state for s in self]
        self.simgr_viewer.simgr.drop(stash=self.stash_name, filter_func=lambda state: state in self.states)
        self.refresh()

    def delete_stash(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        self.simgr_viewer.simgr._stashes.pop(self.stash_name)
        self.simgr_viewer.refresh()

    def paste_states(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        self.simgr_viewer.paste_from_clipboard(self.stash_name)
        self.refresh()

    def move_states(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        self.simgr_viewer.move_to_stash(self.stash_name)
        self.refresh()


class StateTreeItem(SimgrViewerAbstractTreeItem):
    def __init__(self, state, simgr_viewer) -> None:
        self.state = state
        self.simgr_viewer: QSimulationManagerViewer = simgr_viewer
        super().__init__([str(state)])
        self.setData(0, 1, state)

    @property
    def stash_name(self):
        return self.parent().stash_name

    def handle_context_menu_event(self, event) -> None:
        menu = QMenu()
        self.add_menu_action(menu, "Copy state", self.copy_states)
        self.add_menu_action(menu, "Cut state", self.cut_states)
        self.add_menu_action(menu, "Delete state", self.delete_states)
        if self.simgr_viewer.state_clipboard:
            self.add_menu_action(menu, "Paste state", self.paste_states)
        menu.exec_(QCursor.pos())

    def add_menu_action(self, menu, string: str, action) -> None:
        plural = ""
        if len(self.simgr_viewer.selectedItems()) > 1:
            plural = "s"
        menu.addAction(string + plural, action)

    def copy_states(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        self.simgr_viewer.copy_selected_to_clipboard()

    def cut_states(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        self.simgr_viewer.cut_selected_to_clipboard()

    def delete_states(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        self.simgr_viewer.delete_selected_states()

    def paste_states(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        self.simgr_viewer.paste_from_clipboard(self.stash_name)


class QSimulationManagerViewer(QTreeWidget):
    state_clipboard: list[SimState]

    def __init__(self, simgr, parent=None) -> None:
        super().__init__(parent)

        self.setColumnCount(1)
        self.setHeaderHidden(True)

        self.simgr = simgr
        self.state_clipboard = []

        self._init_widgets()

        self.simgr.am_subscribe(self.refresh)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

    def _stash_to_selected_states(self):
        stash_to_states = defaultdict(list)
        for state_tree_item in self.selectedItems():
            stash_to_states[state_tree_item.stash_name].append(state_tree_item.state)
        return stash_to_states

    def copy_selected_to_clipboard(self) -> None:
        self.state_clipboard = [item.state.copy() for item in self.selectedItems()]

    def cut_selected_to_clipboard(self) -> None:
        self.copy_selected_to_clipboard()
        self.delete_selected_states()

    def delete_selected_states(self) -> None:
        stash_to_states = self._stash_to_selected_states()
        for stash_name, states in stash_to_states.items():
            self.simgr.drop(stash=stash_name, filter_func=lambda state, state_set=states: state in state_set)
            self.get_stash_tree_item(stash_name).refresh()

    def paste_from_clipboard(self, stash_name: str) -> None:
        self.simgr.populate(stash_name, self.state_clipboard)
        self.get_stash_tree_item(stash_name).refresh()

    def move_to_stash(self, stash_name: str):
        lambda_str = ""
        lambda_func = None
        while True:
            lambda_str, accepted = QInputDialog.getText(
                self, "Move state from active to here", "Condition lambda", text=lambda_str
            )
            if not accepted:
                return
            try:
                lambda_func = eval(lambda_str)  # pylint: disable=eval-used
                if not isfunction(lambda_func):
                    raise ValueError
            except Exception as e:  # pylint: disable=broad-except
                QMessageBox.critical(self, "Exception!", str(e))
                continue
            break
        self.simgr.move(from_stash="active", to_stash=stash_name, filter_func=lambda_func)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        item = self.itemAt(event.pos())
        if item is not None:
            item.handle_context_menu_event(event)
        else:
            menu = QMenu()
            menu.addAction("Create new stash", self._create_new_stash)
            menu.exec_(QCursor.pos())

    def _create_new_stash(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
        stash_name, accepted = QInputDialog.getText(self, "Stash name", "Blah")

        if not accepted or stash_name.strip() == "":
            # The user didn't provide a stash name
            return

        if stash_name in self.simgr.stashes:
            QMessageBox.critical(
                None,
                "Duplicate stash name",
                f"A stash with the name {stash_name} already exists in the current simulation manager.",
            )
            return
        self.simgr._stashes[stash_name] = []
        self.refresh()

    def refresh(self, **kwargs) -> None:
        if kwargs.get("src", "") != "simgr_viewer":
            self._init_widgets()

    def current_state(self):
        item = self.currentItem()
        if item is None:
            return None
        return item.data(0, 1)

    def select_state(self, state) -> None:
        if state is None:
            self.setCurrentItem(None)
        else:
            for i in range(self.topLevelItemCount()):
                item = self.topLevelItem(i)
                for j in range(item.childCount()):
                    subitem = item.child(j)
                    if subitem.data(0, 1) == state:
                        self.setCurrentItem(subitem)
                        break
                else:
                    continue
                break

    def get_stash_tree_item(self, stash_name: str):
        return self.stash_tree_items[stash_name]

    def _init_widgets(self) -> None:
        # save expanded state
        expended_stash = {}
        selected_state = set()
        for topidx in range(self.topLevelItemCount()):
            top = self.topLevelItem(topidx)
            expended_stash[top.stash_name] = top.isExpanded()
            for idx in range(top.childCount()):
                item = top.child(idx)
                if item.isSelected():
                    selected_state.add(item.state)

        self.clear()

        if self.simgr.am_none:
            return

        self.stash_tree_items = {}
        for stash_name, _stash in self.simgr.stashes.items():  # pylint: disable=unused-variable
            # if not stash and stash_name not in ('active', 'deadended', 'avoided'):
            #     continue
            item = StashTreeItem(stash_name, simgr_viewer=self)
            self.stash_tree_items[stash_name] = item
            self.addTopLevelItem(item)
            item.setExpanded(expended_stash.get(stash_name, False))



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from angr import ailment
from PySide6.QtCore import Qt
from PySide6.QtGui import QTextCharFormat

from angrmanagement.config import Conf
from angrmanagement.utils import string_at_addr

from .base_objects import BlockTreeNode

if TYPE_CHECKING:
    from PySide6.QtWidgets import QGraphicsSceneMouseEvent

    from angrmanagement.data.instance import Instance


class QAilObj(BlockTreeNode):
    """
    Renders an AIL object
    """

    def __init__(
        self, obj: Any, instance: Instance, *args, stmt=None, block_addr=None, block_idx=None, stmt_idx=None, **kwargs
    ) -> None:
        self.stmt = stmt or obj
        self.block_addr = block_addr
        self.block_idx = block_idx
        self.stmt_idx = stmt_idx
        self.instance = instance
        super().__init__(obj, *args, **kwargs)

    def create_subobjs(self, obj: Any) -> None:
        self.add_ailobj(obj)

    def add_ailobj(self, obj: Any) -> None:
        """
        Map appropriate AIL type to the display type
        """
        subobjcls = {
            ailment.statement.Assignment: QAilAssignmentObj,
            ailment.statement.Store: QAilStoreObj,
            ailment.statement.Jump: QAilJumpObj,
            ailment.statement.ConditionalJump: QAilConditionalJumpObj,
            ailment.statement.Return: QAilReturnObj,
            ailment.statement.SideEffectStatement: QAilSideEffectObj,
            ailment.expression.Call: QAilCallObj,
            ailment.expression.Const: QAilConstObj,
            ailment.expression.Tmp: QAilTmpObj,
            ailment.expression.Register: QAilRegisterObj,
            ailment.expression.UnaryOp: QAilUnaryOpObj,
            ailment.expression.BinaryOp: QAilBinaryOpObj,
            ailment.expression.Convert: QAilConvertObj,
            ailment.expression.Load: QAilLoadObj,
        }.get(type(obj), QAilTextObj)
        subobj = subobjcls(obj, self.instance, self.infodock, parent=self, options=self.options, stmt=self.stmt)
        self._add_subobj(subobj)

    @property
    def should_highlight_line(self):
        if self.block_addr is not None:
            stmt_addr = (self.block_addr, self.block_idx, self.stmt_idx)
            if stmt_addr in self.infodock.selected_ail_statements:
                return True
        return super().should_highlight_line

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:  # pylint: disable=unused-argument
        super().mousePressEvent(event)
        button = event.button()
        if button == Qt.MouseButton.LeftButton and self.block_addr is not None:
            stmt_addr = (self.block_addr, self.block_idx, self.stmt_idx)
            self.infodock.select_ail_statement(stmt_addr)


class QAilTextObj(QAilObj):
    """
    Renders an AIL object via __str__
    """

    def create_subobjs(self, obj: Any) -> None:
        self.add_text(str(obj))


class QAilAssignmentObj(QAilTextObj):
    """
    Renders an ailment.statement.Assignment
    """

    def create_subobjs(self, obj: ailment.statement.Assignment) -> None:
        self.add_ailobj(obj.dst)
        self.add_text(" = ")
        self.add_ailobj(obj.src)


class QAilStoreObj(QAilTextObj):
    """
    Renders an ailment.statement.Store
    """

    def create_subobjs(self, obj: ailment.statement.Store) -> None:
        if obj.variable is None or not self.options.show_variables:
            self.add_text("*(")
            self.add_ailobj(obj.addr)
            self.add_text(") = ")
            self.add_ailobj(obj.data)
        else:
            self.add_variable(obj.variable)
            self.add_text(" = ")
            self.add_ailobj(obj.data)


class QAilJumpObj(QAilTextObj):
    """
    Renders an ailment.statement.Jump
    """

    def create_subobjs(self, obj: ailment.statement.Jump) -> None:
        self.add_text("goto ")
        self.add_ailobj(obj.target)


class QAilConditionalJumpObj(QAilTextObj):
    """
    Renders an ailment.statement.ConditionalJump
    """

    def create_subobjs(self, obj: ailment.statement.ConditionalJump) -> None:
        self.add_text("if ")
        self.add_ailobj(obj.condition)

        if self.options.show_conditional_jump_targets:
            self.add_text(" goto ")
            self.add_ailobj(obj.true_target)
            self.add_text(" else goto ")
            self.add_ailobj(obj.false_target)


class QAilReturnObj(QAilTextObj):
    """
    Renders an ailment.statement.Return
    """

    def create_subobjs(self, obj: ailment.statement.Return) -> None:
        self.add_text("return ")
        for expr in obj.ret_exprs:
            self.add_ailobj(expr)


class QAilSideEffectObj(QAilTextObj):
    """
    Renders an ailment.statement.SideEffectStatement, which wraps a Call expression
    and optionally includes a return value destination (ret_expr).
    """

    def create_subobjs(self, obj: ailment.statement.SideEffectStatement) -> None:
        if obj.ret_expr is not None:
            self.add_ailobj(obj.ret_expr)
            self.add_text(" = ")
        self.add_ailobj(obj.expr)


class QAilCallObj(QAilTextObj):
    """
    Renders an ailment.expression.Call
    """

    def create_subobjs(self, obj: ailment.expression.Call) -> None:
        self.add_ailobj(obj.target)
        self.add_text("(")
        if obj.args:
            for i, arg in enumerate(obj.args):
                if i > 0:
                    self.add_text(", ")
                self.add_ailobj(arg)
        self.add_text(")")


class QAilConstObj(QAilTextObj):
    """
    Renders an ailment.expression.Const
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_constant_color)
        return fmt

    def create_subobjs(self, obj: ailment.expression.Const) -> None:
        if not isinstance(obj.value, int):
            return

        # take care of labels first
        kb = self.infodock.disasm_view.disasm.kb
        if obj.value in kb.labels:
            self.add_text(kb.labels[obj.value])
            return

        data_str = string_at_addr(
            self.instance.cfg,
            obj.value,
            self.instance.project,  # type: ignore
        )
        if data_str:
            self.add_text(data_str)
        else:
            self.add_text(f"{obj.value:#x}")

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_block_tree_node, QAilConstObj)
            and self.infodock.selected_block_tree_node.obj.value == self.obj.value
        )

    def mouseDoubleClickEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseDoubleClickEvent(event)
        button = event.button()
        if button == Qt.MouseButton.LeftButton:
            src_ins_addr = getattr(self.stmt, "ins_addr", None)
            self.infodock.disasm_view.jump_to(self.obj.value, src_ins_addr=src_ins_addr, use_animation=True)


class QAilTmpObj(QAilTextObj):
    """
    Renders an ailment.expression.Tmp
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt


class QAilRegisterObj(QAilTextObj):
    """
    Renders an ailment.expression.Register
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt

    def create_subobjs(self, obj: ailment.expression.Register) -> None:
        if obj.variable is not None and self.options.show_variables:
            self.add_variable(obj.variable)
        else:
            s = f"{obj.reg_name}" if hasattr(obj, "reg_name") else f"reg_{obj.reg_offset}<{obj.bits // 8}>"
            self.add_text(s)

    def should_highlight(self) -> bool:
        sel = self.infodock.selected_block_tree_node
        return isinstance(sel, QAilRegisterObj) and sel.obj == self.obj


class QAilUnaryOpObj(QAilTextObj):
    """
    Renders an ailment.expression.UnaryOp
    """

    def create_subobjs(self, obj: ailment.expression.UnaryOp) -> None:
        self.add_text("(")
        self.add_text(obj.op + " ")
        self.add_ailobj(obj.operand)
        self.add_text(")")


class QAilBinaryOpObj(QAilTextObj):
    """
    Renders an ailment.expression.BinaryOp
    """

    def create_subobjs(self, obj: ailment.expression.BinaryOp) -> None:
        self.add_text("(")
        self.add_ailobj(obj.operands[0])
        verbose_op = obj.OPSTR_MAP.get(obj.verbose_op, obj.verbose_op)
        if verbose_op is None:
            verbose_op = "unknown_op"
        self.add_text(" " + verbose_op + " ")
        self.add_ailobj(obj.operands[1])
        self.add_text(")")


class QAilConvertObj(QAilTextObj):
    """
    Renders an ailment.expression.Convert
    """

    def create_subobjs(self, obj: ailment.expression.Convert) -> None:
        self.add_text(f"Conv({obj.from_bits}->{obj.to_bits}, ")
        self.add_ailobj(obj.operand)
        self.add_text(")")


class QAilLoadObj(QAilTextObj):
    """
    Renders an ailment.expression.Load
    """

    def create_subobjs(self, obj: ailment.expression.Load) -> None:
        if obj.variable is not None and self.options.show_variables:
            self.add_variable(obj.variable)
        else:
            self.add_text("*(")
            self.add_ailobj(obj.addr)
            self.add_text(")")



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.analyses.deobfuscator import APIObfuscationFinder, StringObfuscationFinder

from angrmanagement.data.analysis_options import AnalysisConfiguration

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class APIDeobfuscationConfiguration(AnalysisConfiguration):
    """
    Configuration for API deobfuscation.
    """

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "api_deobfuscation"
        self.display_name = "Deobfuscate API usage"
        self.description = "Search for 'obfuscated' API use and attempt to deobfuscate it."
        self.enabled = False


class StringDeobfuscationConfiguration(AnalysisConfiguration):
    """
    Configuration for String deobfuscation.
    """

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "string_deobfuscation"
        self.display_name = "Deobfuscate Strings"
        self.description = "Search for 'obfuscated' strings and attempt to deobfuscate them."
        self.enabled = False


class APIDeobfuscationJob(InstanceJob):
    """
    Job for deobfuscating API usage.
    """

    def __init__(self, instance: Instance, on_finish=None) -> None:
        super().__init__("API Deobfuscation", instance, on_finish=on_finish)

    def run(self, ctx: JobContext) -> None:
        self.instance.project.analyses[APIObfuscationFinder].prep(progress_callback=ctx.set_progress)(
            variable_kb=self.instance.pseudocode_variable_kb
        )

    def __repr__(self) -> str:
        return "APIDeobfuscationJob"


class StringDeobfuscationJob(InstanceJob):
    """
    Job for deobfuscating strings.
    """

    def __init__(self, instance: Instance, on_finish=None) -> None:
        super().__init__("String Deobfuscation", instance, on_finish=on_finish)

    def run(self, ctx: JobContext) -> None:
        self.instance.project.analyses[StringObfuscationFinder].prep(progress_callback=ctx.set_progress)()

    def __repr__(self) -> str:
        return "StringDeobfuscationJob"



from __future__ import annotations

import os

APP_LOCATION = str(os.path.join(os.path.dirname(os.path.realpath(__file__))))
RES_LOCATION = str(os.path.join(APP_LOCATION, "resources"))
IMG_LOCATION = str(os.path.join(RES_LOCATION, "images"))
FONT_LOCATION = str(os.path.join(RES_LOCATION, "fonts"))
THEME_LOCATION = str(os.path.join(RES_LOCATION, "themes"))



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from PySide6.QtWidgets import QFileDialog

from angrmanagement.plugins import BasePlugin

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class SourceImporterPlugin(BasePlugin):
    """
    A plugin that adds scraping source code from the local filesystem and displaying it as an alternative to the
    pseudocode.
    """

    DISPLAY_NAME = "Source Importer"

    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)

        self.source_paths = []
        self._import_from_project()

    def handle_project_initialization(self) -> None:
        self._import_from_project()

    def _import_from_project(self) -> None:
        self.source_paths = []
        if self.workspace.main_instance.original_binary_path:
            self.source_paths.append(os.path.dirname(self.workspace.main_instance.original_binary_path))

    def decompile_callback(self, func) -> None:
        for source_root in self.source_paths:
            self.workspace.main_instance.project.analyses.ImportSourceCode(
                func, flavor="source", source_root=source_root
            )

    MENU_BUTTONS = ["Import source path"]

    def handle_click_menu(self, idx: int) -> None:
        if idx != 0:
            return
        result = QFileDialog.getExistingDirectory(
            self.workspace.main_window,
            "Select source root",
            ".",
            QFileDialog.Option.ShowDirsOnly | QFileDialog.Option.DontResolveSymlinks,
        )
        if result is not None:
            self.source_paths.append(result)



from __future__ import annotations


class FunctionHeader:
    __slots__ = (
        "name",
        "demangled_name",
        "prototype",
        "args",
    )

    def __init__(self, name: str, demangled_name: str | None, prototype=None, args=None) -> None:
        self.name = name
        self.demangled_name = demangled_name
        self.prototype = prototype
        self.args = args


class Variables:
    __slots__ = ["variables"]

    def __init__(self, variables) -> None:
        self.variables = variables


class PhiVariable(Variables):
    __slots__ = ["variable"]

    def __init__(self, variable, variables) -> None:
        super().__init__(variables)
        self.variable = variable


class Label:
    __slots__ = ["addr", "text"]

    def __init__(self, addr: int, text: str) -> None:
        self.addr = addr
        self.text = text



from __future__ import annotations

import logging
import traceback

from angrmanagement.logic.threads import gui_thread_schedule
from angrmanagement.utils.namegen import NameGenerator

log = logging.getLogger(__name__)


class EventSentinel:
    def __init__(self, logging_permitted: bool = True) -> None:
        self.am_subscribers = []
        self.am_logging_permitted: bool = logging_permitted

    def am_subscribe(self, listener) -> None:
        if listener is not None:
            self.am_subscribers.append(listener)

    def am_unsubscribe(self, listener) -> None:
        if listener is not None:
            try:
                self.am_subscribers.remove(listener)
            except ValueError:
                if self.am_logging_permitted:
                    log.warning("Double-unsubscribe of %s from %s", listener, self)
                else:
                    print("Double-unsubscribe of listener")  # No f-string in case str uses logging
                    traceback.print_exc()

    def am_event(self, **kwargs) -> None:
        for listener in self.am_subscribers:
            try:
                listener(**kwargs)
            except Exception:  # pylint: disable=broad-except
                if self.am_logging_permitted:
                    log.exception("Error raised from event of %s", self)
                else:
                    print("Error raised from event")  # No f-string in case str uses logging
                    traceback.print_exc()

    def am_event_gui(self, **kwargs) -> None:
        gui_thread_schedule(self.am_event, kwargs=kwargs)


class ObjectContainer(EventSentinel):
    """
    A proxy for a given object with EventSentinel functionality added on
    Note: While interprocess event notifications are possible via the async_ flag,
    the contents of the shared object are *not* synchronized between processes;
    only the kwargs passed to the am_event of EventSentinel are synchronized
    """

    def __init__(self, obj, name: str | None = None, logging_permitted: bool = True) -> None:
        super().__init__(logging_permitted=logging_permitted)
        self._am_obj = None
        self.am_obj = obj
        self.am_name = name if name is not None else NameGenerator.random_name()

    # cause events to propagate upward through nested objectcontainers
    @property
    def am_obj(self):
        return self._am_obj

    @am_obj.setter
    def am_obj(self, v):
        if v is self:
            raise ValueError("Cannot set am_obj to self")
        if type(self._am_obj) is ObjectContainer:
            self._am_obj.am_unsubscribe(self.__forwarder)
        if type(v) is ObjectContainer:
            v.am_subscribe(self.__forwarder)
        self._am_obj = v

    @property
    def am_none(self) -> bool:
        return self._am_obj is None

    def __forwarder(self, **kwargs) -> None:
        kwargs["forwarded"] = True
        self.am_event(**kwargs)

    def __getattr__(self, item):
        if item.startswith(("am_", "_am_")):
            return object.__getattribute__(self, item)
        return getattr(self._am_obj, item)

    def __setattr__(self, key, value) -> None:
        if key.startswith(("am_", "_am_")):
            return object.__setattr__(self, key, value)
        return setattr(self._am_obj, key, value)

    def __getitem__(self, item):
        return self._am_obj[item]

    def __setitem__(self, key, value) -> None:
        self._am_obj[key] = value

    def __dir__(self):
        return dir(self._am_obj) + list(self.__dict__) + list(EventSentinel.__dict__) + ["am_obj", "am_full"]

    def __iter__(self):
        return iter(self._am_obj)

    def __len__(self) -> int:
        return len(self._am_obj)

    def __eq__(self, other):
        return self is other or self._am_obj == other

    def __ne__(self, other) -> bool:
        return not self == other

    def __repr__(self) -> str:
        return f"(container: {self.am_name}){repr(self._am_obj)}"



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import sys
import unittest

import angr
from common import AngrManagementTestCase, test_location

from angrmanagement.ui.views.proximity_view import ProximityView
from angrmanagement.ui.widgets.qproximitygraph_block import QProximityGraphBlock


class TestProximityView(AngrManagementTestCase):
    def test_proximity_view(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        prox_view = main.workspace._get_or_create_view("proximity", ProximityView)
        prox_view.function = func

        assert prox_view._proximity_graph is not None
        g = prox_view._graph

        assert g is not None and g.number_of_nodes() > 0
        for node in g.nodes():
            assert isinstance(node, QProximityGraphBlock)


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from __future__ import annotations

from .ail_objects import QAilObj
from .base_objects import BlockTreeNode, BlockTreeNodeOptions, QVariableObj
from .disasm_objects import QFunctionHeader
from .vex_objects import QIROpObj

__all__ = [
    "BlockTreeNode",
    "BlockTreeNodeOptions",
    "QAilObj",
    "QIROpObj",
    "QVariableObj",
    "QFunctionHeader",
]



"""
Test cases for XRefDialog.
"""

from __future__ import annotations

import unittest
from unittest.mock import MagicMock

from common import ProjectOpenTestCase

from angrmanagement.ui.dialogs.xref import XRefDialog


class TestXRefDialog(ProjectOpenTestCase):
    """Test XRefDialog functionality."""

    def test_xref_dialog_with_xrefs_manager(self):
        """Test XRefDialog creation with xrefs_manager (address-based xrefs)."""
        dialog = XRefDialog(
            addr=0x1000,
            dst_addr=0x2000,
            xrefs_manager=self.instance.project.kb.xrefs,
            instance=self.instance,
            parent=self.main,
        )

        assert dialog is not None
        assert isinstance(dialog, XRefDialog)
        dialog.close()

    def test_xref_dialog_with_variable_manager(self):
        """Test XRefDialog creation with variable_manager (variable-based xrefs)."""
        mock_variable = MagicMock()
        mock_variable_manager = MagicMock()

        dialog = XRefDialog(
            addr=0x1000,
            variable=mock_variable,
            variable_manager=mock_variable_manager,
            instance=self.instance,
            parent=self.main,
        )

        assert dialog is not None
        assert isinstance(dialog, XRefDialog)
        dialog.close()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import abc
import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING

import angr
import angr.flirt
from angr.flirt import FLIRT_SIGNATURES_BY_ARCH, FlirtSignature
from PySide6.QtWidgets import QFileDialog

from angrmanagement.config import Conf
from angrmanagement.utils.env import app_root, is_pyinstaller

from .object_container import ObjectContainer

if TYPE_CHECKING:
    from .instance import Instance

_l = logging.getLogger(__name__)

#
# FLIRT Signatures
#


def init_flirt_signatures() -> None:
    if Conf.flirt_signatures_root:
        # if it's a relative path, it's relative to the angr-management package
        if os.path.isabs(Conf.flirt_signatures_root):
            flirt_signatures_root = Conf.flirt_signatures_root
        else:
            if is_pyinstaller():
                flirt_signatures_root = os.path.join(app_root(), Conf.flirt_signatures_root)
            else:
                # when running as a Python package, we should use the git submodule, which is on the same level
                # with (instead of inside) the angrmanagement module directory.
                flirt_signatures_root = os.path.join(app_root(), "..", Conf.flirt_signatures_root)
        flirt_signatures_root = os.path.normpath(flirt_signatures_root)
        _l.info("Loading FLIRT signatures from %s.", flirt_signatures_root)
        angr.flirt.load_signatures(flirt_signatures_root)


class Signature(abc.ABC):
    def __init__(self, type_name: str, filename: str, name: str = None):
        self.type_name = type_name
        self.sig_path = filename
        if not name:
            name = Path(self.sig_path).stem
        self.sig_name = name
        self.arch = ""
        self.platform = ""
        self.compiler = ""
        self.os_name = ""

    @abc.abstractmethod
    def apply_signature(self, mgr: SignatureManager, dry_run: bool = True, ignore_addresses: set[int] = None) -> None:
        pass


class PrecomputedSignature(Signature):
    def __init__(self, type_name: str, filename: str, matches: dict[int, str], name: str = None):
        super().__init__(type_name, filename=filename, name=name)
        self.matches = matches

    def apply_signature(self, mgr: SignatureManager, dry_run: bool = True, ignore_addresses: set[int] = None) -> None:
        if ignore_addresses is None:
            ignore_addresses = set()
        if dry_run:
            mgr.dryrun_results[self.sig_path] = self.matches
            return

        for func_addr, name in self.matches.items():
            if ignore_addresses and func_addr in ignore_addresses:
                continue
            func = mgr.instance.project.kb.functions.get_by_addr(func_addr)
            if not func:
                continue
            func.name = name
            func.is_default_name = False
            func.from_signature = "bindiff"


class WrappedFlirtSignature(Signature):
    def __init__(self, flirt: FlirtSignature):
        super().__init__("FLIRT", flirt.sig_path, name=flirt.sig_name)
        self.arch = flirt.arch
        self.platform = flirt.platform
        self.compiler = flirt.compiler
        self.os_name = flirt.os_name
        self.flirt = flirt

    def apply_signature(self, mgr: SignatureManager, dry_run: bool = True, ignore_addresses: set[int] = None) -> None:
        if ignore_addresses is None:
            ignore_addresses = set()
        fl = mgr.instance.project.analyses.Flirt(self.sig_path, dry_run=dry_run or len(ignore_addresses) > 0)
        if not dry_run and not len(ignore_addresses) > 0:
            return

        sig_name, dryrun_results = next(iter(fl.matched_suggestions.values()))
        if dry_run:
            mgr.dryrun_results[self.sig_path] = dryrun_results
            return

        # Manually apply the changes with a reduced set of addresses
        matched_with_ignore = {addr: v for addr, v in dryrun_results.items() if addr not in ignore_addresses}
        _l.info("Applying %s/%s signatures", len(matched_with_ignore), len(dryrun_results))
        fl._apply_changes(
            sig_name if not fl._temporary_sig else None,
            matched_with_ignore,
        )


class SignatureManager:
    """
    Manager of function signatures.
    """

    def __init__(self, instance: Instance) -> None:
        self.signatures: ObjectContainer = ObjectContainer([], "List of function signatures")
        self.dryrun_results: dict[str, dict[int, str]] = {}
        self.instance = instance

    def sync_from_angr(self):
        for _arch, sigs in FLIRT_SIGNATURES_BY_ARCH.items():
            for sig in sigs:
                self.add_signature(WrappedFlirtSignature(sig))

    def clear(self) -> None:
        self.signatures.clear()
        self.signatures.am_event()

    def add_signature(self, sig: Signature) -> None:
        self.signatures.append(sig)
        self.signatures.am_event(added=sig)

    def add_precomputed_signature(self, type_name: str, filename: str, matches: dict[int, str]) -> None:
        sig = PrecomputedSignature(type_name, filename, matches)
        self.dryrun_results[filename] = matches
        self.add_signature(sig)
        return sig

    def apply_signatures(self, sigs: list[Signature], dry_run: bool = True, ignore_addresses: set[int] = None):
        for sig in sigs:
            sig.apply_signature(self, dry_run, ignore_addresses)

    def select_bindiff_base_files(self) -> list[str] | None:
        """
        Open a dialog to select binaries for bindiff analysis.
        Returns the list of selected filenames, or None if cancelled.
        """
        filenames, _ = QFileDialog.getOpenFileNames(
            None,
            "Load binaries to bindiff from",
            "",
            "Object Files (*.a *.o);;All files (*)",
        )
        return filenames if filenames else None

    def load_signatures(self) -> None:
        # open a dialog to select signatures files
        filenames, _ = QFileDialog.getOpenFileNames(
            None,
            "Load FLIRT signature files",
            "",
            "FLIRT signature files (*.sig);;All files (*)",
        )
        for filename in filenames:
            basename = os.path.basename(filename)
            meta_path = filename[: filename.rindex(".")] + ".meta" if "." in basename else None
            r = angr.flirt.load_signature(filename, meta_path=meta_path)
            if r is not None:
                arch, sig = r
                self.add_signature(WrappedFlirtSignature(sig))
            else:
                _l.error("Failed to load signature file %s.", filename)

    def get_match_count(self, sig: Signature) -> int | None:
        if sig.sig_path in self.dryrun_results:
            return len(self.dryrun_results[sig.sig_path])
        return None

    def get_matches(self, sig: Signature) -> dict[int, str] | None:
        """Get the matched functions for a signature as a dict of {address: name}."""
        if sig.sig_path in self.dryrun_results:
            return self.dryrun_results[sig.sig_path]
        return None
